#!/usr/bin/env python
# -*- coding: utf-8 -*-
# TAMkin is a post-processing toolkit for normal mode analysis, thermochemistry
# and reaction kinetics.
# Copyright (C) 2008-2012 Toon Verstraelen <Toon.Verstraelen@UGent.be>, An Ghysels
# <An.Ghysels@UGent.be> and Matthias Vandichel <Matthias.Vandichel@UGent.be>
# Center for Molecular Modeling (CMM), Ghent University, Ghent, Belgium; all
# rights reserved unless otherwise stated.
#
# This file is part of TAMkin.
#
# TAMkin is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# In addition to the regulations of the GNU General Public License,
# publications and communications based in parts on this program or on
# parts of this program are required to cite the following article:
#
# "TAMkin: A Versatile Package for Vibrational Analysis and Chemical Kinetics",
# An Ghysels, Toon Verstraelen, Karen Hemelsoet, Michel Waroquier and Veronique
# Van Speybroeck, Journal of Chemical Information and Modeling, 2010, 50,
# 1736-1750W
# http://dx.doi.org/10.1021/ci100099g
#
# TAMkin is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
#--


from lot_basis import lots_list
import html
from kin import *

from molmod.units import kjmol
from molmod.constants import boltzmann

import matplotlib.pyplot as pt, numpy


# ugly hack
import matplotlib.mathtext
matplotlib.mathtext.Parser._function_names.add("theory")


def get_av(fn_template):
    try:
        gauche_k = load_summary(fn_template % "gauche")[0]
        trans_k = load_summary(fn_template % "trans")[0]
    except IOError:
        return None

    gauche_lnk = numpy.log(gauche_k)
    trans_lnk = numpy.log(trans_k)

    lnav = ((gauche_lnk - experimental_lnk).mean() + (trans_lnk - experimental_lnk).mean())/2
    return numpy.exp(lnav)


def get_rmsd(fn_template):
    try:
        gauche_k = load_summary(fn_template % "gauche")[0]
        trans_k = load_summary(fn_template % "trans")[0]
    except IOError:
        return None

    gauche_lnk = numpy.log(gauche_k)
    trans_lnk = numpy.log(trans_k)

    lnav = numpy.sqrt(((gauche_lnk - experimental_lnk)**2 + (trans_lnk - experimental_lnk)**2).sum()/8)
    return numpy.exp(lnav)


#get_data = get_av
get_data = get_rmsd


def do_plot(data, legend, fn_png):
    pt.clf()
    pt.figure(0, (6,8))
    #pt.axes([left, bottom, width, height])
    # in fractional coordinates
    pt.axes([0.20, 0.070, 0.76, 0.84])
    handles = []
    labels = []
    for key, symbol, label in legend:
        x = []
        y = []
        for index, lot in enumerate(lots_list):
            av = data.get((key, lot.key))
            if av is not None:
                x.append(av)
                y.append(-index)
        handle = pt.plot(x, y, symbol)
        handles.append(handle)
        labels.append(label)
    pos = -numpy.arange(len(lots_list))
    ylabels = [lot.key for lot in lots_list]
    pt.yticks(pos, ylabels, size="small")
    pt.xticks(size="small")
    #pt.xlabel(r"Geometric mean of $k_{\theory}/k_{\exp}$", size="small")
    pt.xlabel(r"$\exp(RMSD(\ln(k_{\theory}) - \ln($k_{\exp})))$", size="small")
    pt.semilogx()
    pt.axvline(0.1, color="k", zorder=-5)
    pt.axvline(1.0, color="k", zorder=-5)
    pt.axvline(10.0, color="k", zorder=-5)
    for y in 0, -5, -10, -15, -20, -25, -30, -35, -40:
        pt.axhline(y, color="k", alpha=0.2, lw=10, zorder=-5)
    if len(legend) > 1:
        legend = pt.figlegend(handles, labels, (0.22,0.915), numpoints=1, handlelength=0, prop={"size":"small"})
        frame = legend.get_frame()
        frame.set_lw(0.0)
    #pt.xlim(1e-8, 1e5)
    pt.xlim(1e0, 1e5)
    pt.ylim(pos[-1]-0.5, 0.5)
    pt.savefig(fn_png)


def figure1():
    #  - comparison of basis-set/geometry for each lot
    #  - remains fixed: HO, BSS
    #  - geometric average (over gauche, trans and temps) of ratio between theory and experiment


    dir_templates = [
        ("low", "%s__6-31gd"),
        ("high", "%s__6-311+g3df2p"),
        ("mixed", "GEO__b3lyp__6-31gd__ENERGY__%s__6-311+g3df2p"),
    ]

    data = {}
    for key, dir_template in dir_templates:
        for lot in lots_list:
            dirname = dir_template % lot.key
            data[(key, lot.key)] = get_data(dirname  + "/ho_bss_summary_%s.txt")

    legend = [
        ("low", "ro", "Consistent, 6-31G(d)"),
        ("high", "gD", "Consistent, 6-311G(3df,2p)"),
        ("mixed", "bs", "2-Component"),
    ]
    do_plot(data, legend, "figure1.png")


def figure2():
    #  - comparison of harmonic oscillator versus inernal rotor for each lot
    #  - remains fixed: BSS, B3LYP geo
    #  - geometric average (over gauche, trans and temps) of ratio between theory and experiment

    data = {}
    for key in "ho", "ir":
        for lot in lots_list:
            dirname = "GEO__b3lyp__6-31gd__ENERGY__%s__6-311+g3df2p" % lot.key
            data[(key, lot.key)] = get_data("%s/%s_bss_summary_%%s.txt" % (dirname, key))

    legend = [
        ("ho", "bs", "2-Component, Harmonic oscillator"),
        ("ir", "yH", "2-Component, 1D Internal rotors"),
    ]
    do_plot(data, legend, "figure2.png")


def figure3():
    #  - comparison of bss and cps for each lot
    #  - remains fixed: HO, B3LYP geo
    #  - geometric average (over gauche, trans and temps) of ratio between theory and experiment

    data = {}
    for key in "bss", "cps":
        for lot in lots_list:
            dirname = "GEO__b3lyp__6-31gd__ENERGY__%s__6-311+g3df2p" % lot.key
            data[(key, lot.key)] = get_data("%s/ho_%s_summary_%%s.txt" % (dirname, key))

    legend = [
        ("bss", "bs", "2-Component, Without correction"),
        ("cps", "m^", "2-Component, Counterpoise corrected barrier"),
    ]
    do_plot(data, legend, "figure3.png")


def figure4():
    #  - final ratios
    #  - remains fixed: IR, CPS, B3LYP geo
    #  - geometric average (over gauche, trans and temps) of ratio between theory and experiment

    data = {}
    for lot in lots_list:
        dirname = "GEO__b3lyp__6-31gd__ENERGY__%s__6-311+g3df2p" % lot.key
        data[("foo", lot.key)] = get_data("%s/ho_bss_summary_%%s.txt" % dirname)
        data[("bar", lot.key)] = get_data("%s/ir_cps_summary_%%s.txt" % dirname)

    legend = [
        ("foo", "bs", "2-Component, Harmonic oscillator, without correction"),
        ("bar", "c*", "2-Component, 1D internal rotors, counterpoise corrected barrier"),
    ]
    do_plot(data, legend, "figure4.png")



if __name__ == "__main__":
    figure1()
    figure2()
    figure3()
    figure4()



# -*- coding: utf-8 -*-
# TAMkin is a post-processing toolkit for normal mode analysis, thermochemistry
# and reaction kinetics.
# Copyright (C) 2008-2012 Toon Verstraelen <Toon.Verstraelen@UGent.be>, An Ghysels
# <An.Ghysels@UGent.be> and Matthias Vandichel <Matthias.Vandichel@UGent.be>
# Center for Molecular Modeling (CMM), Ghent University, Ghent, Belgium; all
# rights reserved unless otherwise stated.
#
# This file is part of TAMkin.
#
# TAMkin is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# In addition to the regulations of the GNU General Public License,
# publications and communications based in parts on this program or on
# parts of this program are required to cite the following article:
#
# "TAMkin: A Versatile Package for Vibrational Analysis and Chemical Kinetics",
# An Ghysels, Toon Verstraelen, Karen Hemelsoet, Michel Waroquier and Veronique
# Van Speybroeck, Journal of Chemical Information and Modeling, 2010, 50,
# 1736-1750W
# http://dx.doi.org/10.1021/ci100099g
#
# TAMkin is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
"""Implementation of free and hindered rotors

   The current implementation supports one-dimensional free and hindered rotors.
   For practical applications it is apparently not necessary to consider higher-
   dimensional hindered rotors. [1]

   [1] Chemical Physics, Vol. 328 (1-3) 251 - 258, 2006
"""

from tamkin.partf import Info, StatFysTerms, helper_vibrations, \
    helpert_vibrations, helpertt_vibrations, helper_levels, helpert_levels, \
    helpertt_levels
from tamkin.nma import NMA, MBH
from tamkin.geom import transrot_basis

from molmod import deg, kjmol, angstrom, centimeter, amu, boltzmann, \
    lightspeed, cached

import numpy as np


__all__ = [
    "HarmonicBasis", "RotorError", "compute_cancel_frequency_mbh",
    "compute_moments", "Rotor"
]


class HarmonicBasis(object):
    """A harmonic basis set for periodic one-dimensional QM systems

       In addition to the definition of the basis set, this class also
       implements the kinetic and potential energy operators required for the
       solution of the schrodinger equation. The workflow is as follows:

       >>> a = 10.0                             # the size of the system
       >>> hb = HarmonicBasis(10, a)            # create basis object
       >>> grid = np.arange(0.0, 10.01, 1.0)    # define a grid
       >>> v = -np.exp(-((grid-5)/2)**2)        # define a potential on the grid
       >>> v_coeffs = hb.fit_fn(grid, v, 10)    # expand the potential in the basis
       >>> mass = 1.0
       >>> energies, wfns = hb.solve(mass, v_coeffs, evecs=True) # solve problem
    """
    def __init__(self, nmax, a):
        """
           Arguments:
            | ``nmax`` -- The maximum index used for the gaussian basis. Hence
                          the the number of basis functions is 2*nmax+1
            | ``a`` -- The length of the periodic system


           The basis functions are:
            | 1/sqrt(a/2)
            | cos(2*pi*x/a)/sqrt(a/2)
            | sin(2*pi*x/a)/sqrt(a/2)
            | ...
            | cos(nmax*2*pi*x/a)/sqrt(a/2)
            | sin(nmax*2*pi*x/a)/sqrt(a/2)

           This is an orthonormal basis. This object has methods to construct a
           Hamiltonian operator in this basis and to transform a function on a
           grid into an expansion in this basis (and back).
        """
        self.nmax = nmax
        self.a = a

    size = property(lambda self: 2*self.nmax+1,
        doc="The size of the basis set. (read-only attribute)")

    def get_empty_op(self):
        """Returns an empty operator (zero)"""
        return np.zeros((self.size, self.size), float)

    def get_hamiltonian_op(self, mass, potential):
        """Returns the Hamiltonian operator for the given mass and potential

           Arguments:
            | ``mass`` -- the mass of the particle
            | ``potential`` -- the expansion coefficients of the potential energy
        """
        op = self.get_empty_op()
        self._add_kinetic_op(op, mass)
        self._add_potential_op(op, potential)
        return op

    def _add_kinetic_op(self, op, mass):
        """Add the kinetic energy to the given operator"""
        factor = (0.5/mass*(2*np.pi/self.a)**2)
        tmp = factor*np.arange(1, self.nmax+1)**2
        diag = op.ravel()[::self.size+1]
        diag[1::2] += tmp
        diag[2::2] += tmp

    def _add_potential_op(self, op, potential):
        """Add the potential energy to the given operator"""
        c = np.zeros(self.nmax+1,float)
        s = np.zeros(self.nmax+1,float)
        c[0] = potential[0]/np.sqrt(self.a)
        c[1:] = potential[1::2]/np.sqrt(self.a)
        s[1:] = potential[2::2]/np.sqrt(self.a)
        op[0,0] += c[0]
        for i0 in xrange(1,self.nmax+1):
            op[2*i0-1,0] += c[i0]
            op[2*i0-0,0] += s[i0]
            op[0,2*i0-1] += c[i0]
            op[0,2*i0-0] += s[i0]
        c[1:] /= np.sqrt(2)
        s[1:] /= np.sqrt(2)
        for i0 in xrange(1,self.nmax+1):
            for i1 in xrange(1,self.nmax+1):
                k = i0+i1
                if k <= self.nmax:
                    op[2*i0-1,2*i1-1] += c[k] # cc
                    op[2*i0-0,2*i1-0] -= c[k] # ss
                    op[2*i0-1,2*i1-0] += s[k] # cs
                    op[2*i0-0,2*i1-1] += s[k] # sc
                k = abs(i0-i1)
                if k <= self.nmax:
                    sign = 2*(i0>i1)-1
                    op[2*i0-1,2*i1-1] += c[k] # cc
                    op[2*i0-0,2*i1-0] += c[k] # ss
                    op[2*i0-1,2*i1-0] -= sign*s[k] # cs
                    op[2*i0-0,2*i1-1] += sign*s[k] # sc

    def solve(self, mass, potential, evecs=False):
        """Return the energies and wavefunctions for the given mass and potential

           Arguments:
            | ``mass`` -- the mass of the particle
            | ``potential`` -- the expansion coefficients of the potential energy

           Optional argument:
            | ``evecs`` -- When True, also the eigenstates are returned.
                           [default=False]
        """
        H = self.get_hamiltonian_op(mass, potential)
        if evecs:
            return np.linalg.eigh(H)
        else:
            return np.linalg.eigvalsh(H)

    def eval_fn(self, grid, coeffs):
        """Evaluate the function represented by coeffs

           Arguments:
            | ``grid`` -- the values at which the function must be evaluated
            | ``coeffs`` -- the expansion coefficients
        """
        result = np.zeros(grid.shape, float) + coeffs[0]/np.sqrt(self.a)
        for i in xrange(self.nmax):
            arg = ((i+1)*2*np.pi/self.a)*grid
            result += (coeffs[2*i+1]/np.sqrt(self.a/2))*np.cos(arg)
            result += (coeffs[2*i+2]/np.sqrt(self.a/2))*np.sin(arg)
        return result

    def eval_deriv(self, grid, coeffs):
        """Evaluate the derivative of function represented by coeffs

           Arguments:
            | ``grid`` -- the values at which the derivative must ben evaluated
            | ``coeffs`` -- the expansion coefficients
        """
        result = np.zeros(grid.shape, float)
        for i in xrange(self.nmax):
            scale = ((i+1)*2*np.pi/self.a)
            arg = scale*grid
            result -= (coeffs[2*i+1]*scale/np.sqrt(self.a/2))*np.sin(arg)
            result += (coeffs[2*i+2]*scale/np.sqrt(self.a/2))*np.cos(arg)
        return result

    def eval_deriv2(self, grid, coeffs):
        """Evaluate the second derivative of function represented by coeffs

           Arguments:
            | ``grid`` -- the values at which the second derivative must be
                          evaluated
            | ``coeffs`` -- the expansion coefficients
        """
        result = np.zeros(grid.shape, float)
        for i in xrange(self.nmax):
            scale = ((i+1)*2*np.pi/self.a)
            arg = scale*grid
            result -= (coeffs[2*i+1]*scale**2/np.sqrt(self.a/2))*np.cos(arg)
            result -= (coeffs[2*i+2]*scale**2/np.sqrt(self.a/2))*np.sin(arg)
        return result

    def fit_fn(self, grid, v, dofmax, rotsym=1, even=False, rcond=0.0, v_threshold=0.01):
        """Fit the expansion coefficients that represent function f

           Arguments:
            | ``grid`` -- The x values on which the function f is known.
            | ``v`` -- The function to be represented by expansion coefficients.
            | ``dofmax`` -- The maximum number of cosines in the fit. When
                            even==False, the same number of sines is also
                            included.

           Optional arguments:
            | ``rotsym`` -- Impose this rotational symmetry [default=1]. Must be
                            an integer and is least 1..
            | ``even`` -- Only fit even functions, i.e. cosines. [default=False]
            | ``rcond`` -- The cutoff for the singular values in the least
                           squares fit. [default=1e-10]
            | ``v_threshold`` -- Tolerance on the relative error between the
                                 Fourier expansion and the data points of the
                                 scan. [default=0.01]. Absolute errors smaller
                                 than 1 kJ/mol are always ignored.

           In case the Fourier expansion represents a poor fit (determined by
           v_threshold), a ValueError is raised. It means that you have to
           check your torsional scan datapoints for errors.
        """
        if rotsym < 1:
            raise ValueError("rotsym must be at least 1.")

        # construct the design matrix
        ncos = min(dofmax, self.nmax/rotsym)
        if even:
            A = np.zeros((len(grid), ncos+1), float)
        else:
            A = np.zeros((len(grid), 2*ncos+1), float)
        A[:,0] = 1.0/np.sqrt(self.a)
        counter = 1
        for i in xrange(ncos):
            arg = ((i+1)*rotsym*2*np.pi/self.a)*grid
            A[:,counter] = np.cos(arg)/np.sqrt(self.a/2)
            counter += 1
            if not even:
                A[:,counter] = np.sin(arg)/np.sqrt(self.a/2)
                counter += 1

        coeffs, residuals, rank, S = np.linalg.lstsq(A, v, rcond)

        # check the error
        residual = np.dot(A, coeffs) - v
        rmsd = (residual**2).mean()**0.5
        rms = (v**2).mean()**0.5
        abs_threshold = max(1*kjmol, rms*v_threshold)
        if (rmsd > abs_threshold).any():
            raise ValueError("Residual is too large. (poor Fourier expansion.) rmsd [kJ/mol] = %f, rms [kJ/mol] = %f" % (rmsd/kjmol, rms/kjmol))

        # collect the parameters in a convenient array
        result = np.zeros(self.size)
        result[0] = coeffs[0]
        if even:
            tmp = result[2*rotsym-1::2*rotsym]
            tmp[:ncos] = coeffs[1:]
        else:
            tmp = result[2*rotsym-1::2*rotsym]
            tmp[:ncos] = coeffs[1::2]
            tmp = result[2*rotsym-0::2*rotsym]
            tmp[:ncos] = coeffs[2::2]
        return result


class RotorError(Exception):
    """This exception is raised when :class:`Rotor` specific errors are
       encountered.
    """
    pass


def compute_cancel_frequency_mbh(molecule, dihedral, top_indexes):
    """Compute the frequency of the rotor in the HO approximation

       This function is based on the MBH method and returns the frequency that
       has to be canceled when this mode is replaced by a free or hindered
       rotor.

       Arguments:
        | ``molecule`` -- A Molecule object. (see :mod:`tamkin.data`)
        | ``top_indexes`` -- The indexes of the rotor atoms.
    """
    axis = tuple(dihedral[1:3])
    other_top_indexes = tuple(set(xrange(molecule.size)) - set(top_indexes) - set(axis))
    blocks = [
        axis + tuple(top_indexes),
        axis + tuple(other_top_indexes),
    ]
    nma = NMA(molecule, MBH(blocks))
    if len(nma.freqs) != 7:
        raise RotorError("Expecting 7 frequencies, got %i" % len(nma.freqs))
    non_zero = [i for i in xrange(7) if i not in nma.zeros][0]
    return nma.freqs[non_zero]


def compute_moments(coordinates, masses3, center, axis, indexes):
    """Computes the absolute and the relative moment of an internal rotor

       Arguments:
        | ``coordinates`` -- The coordinates of all atoms, float numpy array
                             with shape (N,3).
        | ``masses3`` -- The diagonal of the mass matrix, each mass is repeated
                         three tines, float numpy array with shape 3N.
        | ``center`` -- A point on the rotation axis. Float numpy array with
                        shape 3.
        | ``axis`` -- A unit vector with the direction of the rotation axis.
                      Float numpy array with shape 3.
        | ``indexes`` -- The indexes of the atoms that belong to the rotor.

       The implementation is based on the transformation of the mass-matrix to
       the internal rotation coordinate. The derivative of the internal
       coordinate towards cartesian coordinates represents a small displacement
       of the atoms. This displacement is constrained to show no external
       linear or rotational moment.
    """
    # the derivative of the cartesian coordinates towards the rotation
    # angle of the top:
    rot_tangent = np.zeros((len(coordinates)*3), float)
    for i in indexes:
        rot_tangent[3*i:3*i+3] = np.cross(coordinates[i]-center, axis)
    # transform this to a derivative without global linear or angular
    # momentum. This is done by projecting on the basis of external degrees
    # of freedom in mass-weighted coordinates, and subsequently subtracting
    # that projection from the original tangent
    basis = transrot_basis(coordinates)
    A = np.dot(basis*masses3, basis.transpose())
    B = np.dot(basis*masses3, rot_tangent)
    alphas = np.linalg.solve(A,B)
    rot_tangent_relative = rot_tangent - np.dot(alphas, basis)
    return (
        (rot_tangent**2*masses3).sum(),
        (rot_tangent_relative**2*masses3).sum()
    )


class Rotor(Info, StatFysTerms):
    """Partition function term for a one-dimensional rotor

       The contribution from the free or hindered rotor to the partition
       function is based on the quantum mechanical solution of the rotational
       motion. To avoid double counting problems, one must also provide the
       frequency of this motion as if it was treated as a harmonic oscillator.
       The corresponding contribution to the partition function is subtracted.
       (Use compute_cancel_frequency to obtain this frequency.)
    """
    def __init__(self, rot_scan, molecule=None, cancel_freq='mbh', suffix=None,
                 rotsym=1, even=False, num_levels=50, dofmax=5,
                 v_threshold=0.01, large_fixed=False):
        """
           Arguments:
            | ``rot_scan`` -- A rotational scan object. (free or hindered rotor
                              information)

           Optional arguments:
            | ``molecule`` -- Molecule to which the rotor applies, is used to
                              compute the cancelation frequency. Not required
                              when the cancel_freq argument is present.
            | ``cancel_freq`` -- The frequency to cancel in the vibrational
                                 partition function. This can also be 'mbh' or
                                 'scan' to indicate that the cancel frequency
                                 should be computed using the MBH method or
                                 based on the second order derivative of the
                                 rotational potential. Note that the latter
                                 option is only possible in the case of the
                                 hindered rotor formalism. [default='mbh']
            | ``suffix`` -- A name suffix used to distinguish between different
                            rotors.
            | ``rotsym`` -- The rotational symmetry of the rotor. [default=1]
            | ``even`` -- True of the rotor is not chiral, i.e. when it has an
                          even potential
            | ``num_levels`` -- The number of energy levels considered in the
                                QM treatment of the rotor [default=50]
            | ``dofmax`` -- The maximum number of cosines used to represent the
                            torsional potential. if the potential is not even,
                            the same number of sines is also used. [default=5]
            | ``v_threshold`` -- Tolerance on the relative error between the
                                 Fourier expansion and the data points of the
                                 scan. [default=0.01]. Absolute errors smaller
                                 than 1 kJ/mol are always ignored.
            | ``large_fixed`` -- When True, assume that the large part of the
                                 system is fixed in space while the small part
                                 rotates. (this means that the absolute moment
                                 of the rotor is used instead of the relative
                                 moment)

           In case the Fourier expansion of the potential represents a poor fit
           (determined by v_threshold), a ValueError is raised. It means that
           you have to check your torsional scan datapoints for errors.
        """
        self.rot_scan = rot_scan
        self.cancel_freq = cancel_freq
        self.cancel_method = 'given by the user'
        # the cancelation frequency
        if self.cancel_freq == 'mbh':
            self.cancel_freq = compute_cancel_frequency_mbh(
                molecule, self.rot_scan.dihedral, self.rot_scan.top_indexes
            )
            self.cancel_method = 'computed with mbh'
        elif self.cancel_freq == 'scan':
            if self.rot_scan.potential is None:
                raise RotorError("The option cancel_freq='scan' can only be used with hindered rotors.")
            self.cancel_method = 'derived from the torsional potential'
            # the actual computation is performed later
        else:
            raise ValueError("Could not interpret cancel_freq=%s" % cancel_freq)
        #
        if suffix is None:
            self.suffix = "_".join(str(i) for i in rot_scan.top_indexes)
        else:
            self.suffix = suffix
        self.rotsym = rotsym
        self.even = even
        self.num_levels = num_levels
        self.dofmax = dofmax
        self.v_threshold = v_threshold
        self.large_fixed = large_fixed
        if rot_scan.potential is None:
            Info.__init__(self, "free_rotor_%s" % self.suffix)
        else:
            Info.__init__(self, "hindered_rotor_%s" % self.suffix)
        StatFysTerms.__init__(self, 2) # two terms

    def init_part_fun(self, nma, partf):
        """See :meth:`tamkin.partf.StatFys.init_part_fun`"""
        if nma.periodic:
            raise NotImplementedError("Rotors in periodic systems are not supported yet")

        self.center = nma.coordinates[self.rot_scan.dihedral[1]]
        self.axis = nma.coordinates[self.rot_scan.dihedral[2]] - self.center
        self.axis /= np.linalg.norm(self.axis)
        self.moment, self.reduced_moment = compute_moments(
            nma.coordinates, nma.masses3, self.center, self.axis, self.rot_scan.top_indexes
        )
        if self.large_fixed:
            moment = self.moment
        else:
            moment = self.reduced_moment
        from molmod.ic import dihed_angle
        self.nma_angle = dihed_angle(nma.coordinates[self.rot_scan.dihedral])[0]
        # the energy levels
        if self.rot_scan.potential is None:
            # free rotor
            self.energy_levels = np.zeros(self.num_levels, float)
            for i in xrange(self.num_levels-1):
                index = i/2+1
                self.energy_levels[i+1] = index**2/(2*moment)
            self.hb = None
            self.v_coeffs = None
            self.v_ref = 0.0
        else:
            # hindered rotor
            self.hb = HarmonicBasis(self.num_levels, 2*np.pi)
            angles, energies = self.potential

            self.v_coeffs = self.hb.fit_fn(angles, energies, self.dofmax,
                self.rotsym, self.even, self.v_threshold)
            self.energy_levels = self.hb.solve(moment, self.v_coeffs)
            self.energy_levels = self.energy_levels[:self.num_levels]

        # the cancelation frequency based on the scan
        if self.hb is None:
            if not isinstance(self.cancel_freq, float):
                raise ValueError('No cancelation frequency was computed for rotor "%s"' % self.name)
        else:
            force_constant = self.hb.eval_deriv2(np.array([self.nma_angle]), self.v_coeffs)[0]
            scan_cancel_freq = np.sqrt(force_constant/moment)/(2*np.pi)
            if self.cancel_freq == 'scan':
                self.cancel_freq = scan_cancel_freq
            elif abs(self.cancel_freq - scan_cancel_freq) > 0.3*abs(self.cancel_freq):
                print 'WARNING: The cancelation frequency of rotor "%s" obtained with MBH (%.1f cm^-1) deviates a lot from the one derived from the scan (%.1f cm^-1).' % (
                    self.name, self.cancel_freq/(lightspeed/centimeter), scan_cancel_freq/(lightspeed/centimeter)
                )

        # print a warning is the cancelation frequency is rather high.
        if self.cancel_freq > 500*(lightspeed/centimeter):
            print 'WARNING: the cancelation frequency of rotor "%s" is rather high: %.1f cm^-1.' % (
                self.name, self.cancel_freq/(lightspeed/centimeter)
            )
        elif self.cancel_freq <= 0:
            print 'WARNING: the cancelation frequency of rotor "%s" is negative: %.1f cm^-1.' % (
                self.name, self.cancel_freq/(lightspeed/centimeter)
            )


        # scaling factors
        self.freq_scaling = partf.vibrational.freq_scaling
        self.zp_scaling = partf.vibrational.zp_scaling
        self.classical = partf.vibrational.classical

    @cached
    def potential(self):
        """A tuple with angles and potential energies (hindered only)

           If the rotor is free, the result is None. The reference for the
           potential energy is the energy of the reference geometry used in the
           partition function.
        """
        if self.rot_scan.potential is None:
            return None
        else:
            a = 2*np.pi
            angles, energies = self.rot_scan.potential.copy()
            # apply periodic boundary conditions
            angles -= np.floor(angles/a)*a
            # set reference energy, which is take to be the energy of the geometry
            # with a dihedral angle closest to the dihedral angle of the reference
            # geometry in the nma object. We can not take the nma energy because
            # the rotational energy barriers may be computed at another level
            # that the nma energy.
            deltas = angles - self.nma_angle
            # apply periodic boundary conditions
            deltas -= np.floor(deltas/a)*a
            # get the right energy
            energies -= energies[deltas.argmin()]
            return angles, energies

    def dump(self, f):
        """Write all the information about the rotor to a file

           This method is part of the PartFun API and should never be called
           directly. It will only work properly once the init_part_fun method
           is called.

           Arguments:
            | ``f`` -- A file-like object.
        """
        Info.dump(self, f)
        # parameters
        print >> f, "    Dihedral indexes: %s" % " ".join(str(i) for i in self.rot_scan.dihedral)
        print >> f, "    Top indexes: %s" % " ".join(str(i) for i in self.rot_scan.top_indexes)
        print >> f, "    Rotational symmetry: %i" % self.rotsym
        print >> f, "    Even potential: %s" % self.even
        if self.rot_scan.potential is None:
            print >> f, "    This is a free rotor"
        else:
            angles, energies = self.rot_scan.potential
            print >> f, "    This is a hindered rotor"
            print >> f, "    Maximum number of cosines in the fit: %i" % self.dofmax
            angles, energies = self.potential
            fit_energies = self.hb.eval_fn(angles, self.v_coeffs)
            rmsd = ((fit_energies - energies)**2).mean()**0.5
            rms = (energies**2).mean()**0.5
            rrmsd = rmsd/rms
            print >> f, "    RMSD of the fit [kJ/mol]: %.2f" % (rmsd/kjmol)
            print >> f, "    Relative RMSD of the fit the fit [%%]: %.1f" % (rrmsd*100)
            print >> f, "    Pearson R^2 of the fit [%%]: %.3f" % ((1-rrmsd**2)*100)

            print >> f, "    Potential: Angle [deg]    Energy [kJ/mol]"
            for i in xrange(len(angles)):
                print >> f, "              % 7.2f         %6.1f" % (angles[i]/deg, energies[i]/kjmol)
        print >> f, "    Number of QM energy levels: %i" % self.num_levels
        # derived quantities
        print >> f, "    Center [A]: % 8.2f % 8.2f % 8.2f" % tuple(self.center/angstrom)
        print >> f, "    Axis [1]: % 8.2f % 8.2f % 8.2f" % tuple(self.axis)
        print >> f, "    Moment [amu*bohr**2]: %f" % (self.moment/amu)
        print >> f, "    Reduced moment [amu*bohr**2]: %f" % (self.reduced_moment/amu)
        print >> f, "    Cancel wavenumber [1/cm]: %.1f" % (self.cancel_freq/(lightspeed/centimeter))
        print >> f, "    The cancelation wavenumber is %s." % self.cancel_method
        self.dump_values(f, "Energy levels [kJ/mol]", self.energy_levels/kjmol, "% 8.2f", 8)
        if self.hb is not None:
            print >> f, "    Number of basis functions: %i" % (self.hb.size)
        print >> f, "    Zero-point contribution [kJ/mol]: %.7f" % (self.free_energy(0.0)/kjmol)

    def plot_levels(self, prefix, temp, do_levels=True, do_data=True):
        """Plots the potential with the energy levels

           Arguments:
            | ``prefix`` -- A filename prefix for the png files.
            | ``temp`` -- A temperature that is used to indicate the statistical
                          weight of each level in the plots

           Optional argument:
            | ``do_levels`` -- When True, the energy levels are plotted.
                               [default=True]
            | ``do_data`` -- When True, the data points are plotted.
                             [default=True]

           One image will be generated:
            | ``${prefix}.png`` -- The potential and the energy levels.
        """
        import matplotlib.pyplot as pt
        pt.clf()
        if do_data:
            # plot the original potential data
            if self.rot_scan.potential is not None:
                angles, energies = self.potential
                pt.plot(angles/deg, energies/kjmol, "rx", mew=2)
            pt.axvline(self.nma_angle/deg, color="silver")
        # plot the fitted potential
        if self.hb is not None:
            step = 0.001
            x = np.arange(0.0, 2*np.pi*(1+0.5*step), 2*np.pi*step)
            v = self.hb.eval_fn(x, self.v_coeffs)
            pt.plot(x/deg, v/kjmol, "k-", linewidth=2)
        if do_levels:
            # plot the energy levels
            eks = self.energy_levels/(temp*boltzmann)
            bfs = np.exp(-eks)
            bfs /= bfs.sum()
            for i in xrange(self.num_levels):
                e = (self.energy_levels[i])/kjmol
                pt.axhline(e, color="b", linewidth=0.5)
                pt.axhline(e, xmax=bfs[i], color="b", linewidth=2)
        if self.rot_scan.potential is not None:
            angles, energies = self.potential
            pt.ylim(
                energies.min()/kjmol,
                1.5*energies.max()/kjmol
            )
            fit_energies = self.hb.eval_fn(angles, self.v_coeffs)
            rmsd = ((fit_energies - energies)**2).mean()**0.5
            rms = (energies**2).mean()**0.5
            rrmsd = rmsd/rms
            title = 'RMSD [kJ/mol] = %.1f    RRMSD [%%] = %.0f    dofmax = %i   rotsym = %i' % (
                rmsd, rrmsd*100, self.dofmax, self.rotsym
            )
            if self.even:
                title += '   even'
            pt.title(title)
        pt.xlim(0, 360)
        pt.ylabel("Energy [kJ/mol]")
        pt.xlabel("Dihedral angle [deg]")
        pt.savefig(prefix)

    def helper_terms(self, temp, n):
        """See :meth:`tamkin.partf.StatFysTerms.helper_terms`"""
        return np.array([
            -helper_vibrations(temp, n, self.cancel_freq, self.classical,
                                 self.freq_scaling, self.zp_scaling),
            helper_levels(temp, n, self.energy_levels, check=True) - temp**n*np.log(self.rotsym),
        ])

    def helpert_terms(self, temp, n):
        """See :meth:`tamkin.partf.StatFysTerms.helpert_terms`"""
        return np.array([
            -helpert_vibrations(temp, n, self.cancel_freq, self.classical,
                                  self.freq_scaling, self.zp_scaling),
            helpert_levels(temp, n, self.energy_levels, check=True),
        ])

    def helpertt_terms(self, temp, n):
        """See :meth:`tamkin.partf.StatFysTerms.helpertt_terms`"""
        return np.array([
            -helpertt_vibrations(temp, n, self.cancel_freq, self.classical,
                                   self.freq_scaling, self.zp_scaling),
            helpertt_levels(temp, n, self.energy_levels, check=True),
        ])



# -*- coding: utf-8 -*-
# TAMkin is a post-processing toolkit for normal mode analysis, thermochemistry
# and reaction kinetics.
# Copyright (C) 2008-2012 Toon Verstraelen <Toon.Verstraelen@UGent.be>, An Ghysels
# <An.Ghysels@UGent.be> and Matthias Vandichel <Matthias.Vandichel@UGent.be>
# Center for Molecular Modeling (CMM), Ghent University, Ghent, Belgium; all
# rights reserved unless otherwise stated.
#
# This file is part of TAMkin.
#
# TAMkin is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# In addition to the regulations of the GNU General Public License,
# publications and communications based in parts on this program or on
# parts of this program are required to cite the following article:
#
# "TAMkin: A Versatile Package for Vibrational Analysis and Chemical Kinetics",
# An Ghysels, Toon Verstraelen, Karen Hemelsoet, Michel Waroquier and Veronique
# Van Speybroeck, Journal of Chemical Information and Modeling, 2010, 50,
# 1736-1750W
# http://dx.doi.org/10.1021/ci100099g
#
# TAMkin is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --


import numpy as np
import pkg_resources
import unittest

from molmod.units import kjmol, centimeter
from molmod.constants import lightspeed

from tamkin import *


__all__ = ["TunnelingTestCase"]


class TunnelingTestCase(unittest.TestCase):
    def test_eckart_leo(self):
        # Data is taken from:
        #   Molecular Physics, May 2003, Vol 1.1, No 9, p1329-1338
        #   M.L. Coote, M.A. Collins, L. Radom
        temp = 298
        forward = np.array([71.7, 49.7, 62.0, 44.5,  6.8], float)*kjmol
        reverse = np.array([71.7, 86.8, 76.0, 78.2, 26.4], float)*kjmol

        frequencies = np.array([
            [3489, 3376, 3592, 3338, 3418],
            [3411, 3315, 3529, 3274, 3376],
            [3398, 3309, 3528, 3257, 3389],
            [2558, 2578, 2611, 2421, 2096],
            [2518, 2537, 2572, 2382, 2070],
            [2530, 2545, 2585, 2397, 2110],
            [1741, 1658, 1760, 1368,  796],
            [1713, 1640, 1739, 1358,  832],
            [1725, 1656, 1770, 1392,  926],
            [1657, 1592, 1673, 1311,  737],
            [1638, 1568, 1649, 1286,  674],
            [1638, 1579, 1659, 1306,  790],
            [1657, 1602, 1697, 1352,  981],
            [1773, 1739, 1824, 1456,  827],
            [2134, 2061, 2197, 1849, 1358],
            [1959, 1865, 2050, 1600, 1221],
            [1941, 1829, 2038, 1569, 1185],
            [2029, 2010, 2073, 1765, 1126],
            [1866, 1859, 1924, 1574, 1046],
        ], float).transpose()*lightspeed/centimeter

        eckart_corrections = np.array([
            [7.8E+5, 3.2E+4, 4.0E+5, 1.1E+4, 7.2E+0],
            [5.9E+5, 2.8E+4, 3.3E+5, 1.0E+4, 7.1E+0],
            [5.6E+5, 2.7E+4, 3.3E+5, 9.8E+3, 7.1E+0],
            [1.2E+4, 2.6E+3, 9.5E+3, 8.0E+2, 5.0E+0],
            [1.0E+4, 2.2E+3, 7.9E+3, 6.9E+2, 4.9E+0],
            [1.1E+4, 2.3E+3, 8.4E+3, 7.3E+2, 5.0E+0],
            [9.7E+1, 4.1E+1, 9.4E+1, 1.0E+1, 1.8E+0],
            [8.2E+1, 3.7E+1, 8.3E+1, 9.8E+0, 1.8E+0],
            [8.8E+1, 4.0E+1, 9.9E+1, 1.1E+1, 2.0E+0],
            [5.9E+1, 3.0E+1, 5.7E+1, 8.2E+0, 1.7E+0],
            [5.2E+1, 2.7E+1, 5.0E+1, 7.5E+0, 1.5E+0],
            [5.2E+1, 2.8E+1, 5.3E+1, 8.1E+0, 1.8E+0],
            [5.9E+1, 3.1E+1, 6.5E+1, 9.6E+0, 2.0E+0],
            [1.2E+2, 6.0E+1, 1.4E+2, 1.4E+1, 1.8E+0],
            [1.1E+3, 2.9E+2, 1.1E+3, 7.8E+1, 3.1E+0],
            [3.7E+2, 1.1E+2, 5.0E+2, 2.6E+1, 2.8E+0],
            [3.3E+2, 9.3E+1, 4.7E+2, 2.3E+1, 2.7E+0],
            [5.8E+2, 2.2E+2, 5.7E+2, 5.4E+1, 2.5E+0],
            [2.1E+2, 1.1E+2, 2.4E+2, 2.4E+1, 2.3E+0],
        ], float).transpose()

        for Ef, Er, fs, cs in zip(forward, reverse, frequencies, eckart_corrections):
            for f, c in zip(fs, cs):
                correction = Eckart._from_parameters(Ef, Er, f)
                our = correction([temp])[0]
                self.assert_(abs(our-c)/c < 0.09)
                #if not (("%.1e" % our)==("%.1e" % c)):
                #    print "%.1e" % our, "%.1e" % c

    def test_eckart_init(self):
        fixed_atoms = load_fixed_g03com(
            pkg_resources.resource_filename(__name__, "../data/test/mat/Zp_p_react.14mei.com"))
        mol_react = load_molecule_g03fchk(
            pkg_resources.resource_filename(__name__, "../data/test/mat/Zp_p_react.28aug.fchk"))
        mol_trans = load_molecule_g03fchk(
            pkg_resources.resource_filename(__name__, "../data/test/mat/Zp_p_TS.28aug.fchk"))

        pf_react = PartFun(NMA(mol_react, PHVA(fixed_atoms)))
        pf_trans = PartFun(NMA(mol_trans, PHVA(fixed_atoms)))

        Eckart([pf_react], pf_trans, [pf_react])

    def test_eckart_tugba(self):
        mol_react1 = load_molecule_g03fchk(
            pkg_resources.resource_filename(__name__, "../data/test/tugba/monomer.fchk"))
        mol_react2 = load_molecule_g03fchk(
            pkg_resources.resource_filename(__name__, "../data/test/tugba/radical.fchk"))
        mol_trans = load_molecule_g03fchk(
            pkg_resources.resource_filename(__name__, "../data/test/tugba/ts.fchk"))
        mol_prod = load_molecule_g03fchk(
            pkg_resources.resource_filename(__name__, "../data/test/tugba/prod.fchk"))

        pf_react1 = PartFun(NMA(mol_react1))
        pf_react2 = PartFun(NMA(mol_react2))
        pf_trans = PartFun(NMA(mol_trans))
        pf_prod = PartFun(NMA(mol_prod))

        eckart = Eckart([pf_react1, pf_react2], pf_trans, [pf_prod])
        eckart(np.array([620,770]))

    def test_miller_tugba(self):
        mol_trans = load_molecule_g03fchk(
            pkg_resources.resource_filename(__name__, "../data/test/tugba/ts.fchk"))
        pf_trans = PartFun(NMA(mol_trans))
        miller = Miller(pf_trans)
        miller(np.array([620,770]))



# -*- coding: utf-8 -*-
# TAMkin is a post-processing toolkit for normal mode analysis, thermochemistry
# and reaction kinetics.
# Copyright (C) 2008-2012 Toon Verstraelen <Toon.Verstraelen@UGent.be>, An Ghysels
# <An.Ghysels@UGent.be> and Matthias Vandichel <Matthias.Vandichel@UGent.be>
# Center for Molecular Modeling (CMM), Ghent University, Ghent, Belgium; all
# rights reserved unless otherwise stated.
#
# This file is part of TAMkin.
#
# TAMkin is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# In addition to the regulations of the GNU General Public License,
# publications and communications based in parts on this program or on
# parts of this program are required to cite the following article:
#
# "TAMkin: A Versatile Package for Vibrational Analysis and Chemical Kinetics",
# An Ghysels, Toon Verstraelen, Karen Hemelsoet, Michel Waroquier and Veronique
# Van Speybroeck, Journal of Chemical Information and Modeling, 2010, 50,
# 1736-1750W
# http://dx.doi.org/10.1021/ci100099g
#
# TAMkin is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
"""Tools for further investigation of a normal mode analysis"""

from tamkin.data import Molecule
from tamkin.nma import NMA
from tamkin.io.charmm import load_peptide_info_charmm

from molmod import lightspeed, angstrom, amu, centimeter

import numpy as np


__all__ = [
    "compute_overlap", "write_overlap",
    "compute_delta", "compute_sensitivity_freq",
    "create_blocks_peptide_charmm", "create_subs_peptide_charmm",
    "plot_spectrum_lines", "plot_spectrum_dos",
    "create_enm_molecule",
]


invcm = lightspeed/centimeter


def compute_overlap(nma1, nma2, filename=None, unit="au"):
    """Compute overlap of modes and print to file if requested

       Arguments:
         | nma1  --  modes and frequencies (see below)
         | nma2  --  modes and frequencies (see below)

       Optional argument:
         | filename  --  when given, the overlap is written to file by the
                       function write_overlap
         | unit  --  unit in which frequencies should be printed in the
                     file: au [default] or 1/centimeter [cm1]

       The nma arguments can have different formats:

       1) an NMA object
       2) a tuple or list with two elements: modes and frequencies
       3) a numpy array with the mass-weighted modes
       4) a numpy array with one mass-weighted mode
    """

    def parse_nma(nma):
        if isinstance(nma, NMA):
            # NMA object
            return nma.modes, nma.freqs
        elif hasattr(nma, "__len__") and len(nma) == 2 and not isinstance(nma, np.ndarray):
            # [modes,freqs] or (modes,freqs)
            return nma
        elif isinstance(nma, np.ndarray) and len(nma.shape) == 2:
            # modes only
            return nma, np.zeros(nma.shape[1], float)
        elif isinstance(nma, np.ndarray) and len(nma.shape) == 1:
            # one mode only
            return nma.reshape(-1,1), np.zeros(1, float)
        else:
            raise TypeError("nma argument has wrong type")

    modes1, freqs1 = parse_nma(nma1)
    modes2, freqs2 = parse_nma(nma2)

    # check dimensions
    if modes1.shape[0] != modes2.shape[0] :
        raise ValueError("Length of columns in modes1 and modes2 should be equal, but found %i and %i." % (modes1.shape[0], modes2.shape[0]))
    # compute overlap
    overlap = np.dot(np.transpose(modes1), modes2)
    if filename is not None:
        write_overlap(freqs1, freqs2, overlap, filename=filename, unit=unit)
    return overlap


def write_overlap(freqs1, freqs2, overlap, filename="overlap.csv", unit="au"):
    """Write the overlap matrix to a csv file

       Arguments:
        | freqs1  --  the list of frequencies associated with the rows of the
                     overlap matrix
        | freqs2  --  the list of frequencies associated with the columns of the
                     overlap matrix
        | overlap  --  the overlap matrix

       Optional arguments:
        | filename  --  the file to write to [default="overlap.csv"]
        | unit      --  unit in which frequencies are printed in file:
                        1/centimeter (cm1) or au  [default]

       The table contains the following blocks:

       +----------+--------------+
       |          | freqs2       |
       +----------+--------------+
       | freqs1^T | mat1^T . mat2|
       +----------+--------------+
    """
    #freqs1 = freqs1 / invcm
    #freqs2 = freqs2 / invcm

    to_append="w+"   # not append, just overwrite
    f = file(filename,to_append)

    [rows,cols] = overlap.shape

    if unit is "au":
        # 1. row of freqs2
        print >> f, ";"+";".join(str(g) for g in freqs2)  #this is the same

        # 2. start each row with freq of freqs1 and continue with overlaps
        for r in range(rows):
            print >> f, str(freqs1[r])+";"+";".join(str(g) for g in overlap[r,:].tolist())

    elif unit is "cm1":
        # 1. row of freqs2
        print >> f, ";"+";".join(str(g*centimeter/lightspeed) for g in freqs2)  #this is the same

        # 2. start each row with freq of freqs1 and continue with overlaps
        for r in range(rows):
            print >> f, str(freqs1[r]*centimeter/lightspeed)+";"+";".join(str(g) for g in overlap[r,:].tolist())

    else:
        raise NotImplementedError("this unit is not implemented/recognized")
    f.close()


def compute_delta(coor1, coor2, masses=None, normalize=False):
    """Compute mass weighted delta vector between two conformations

       It is assumed that the structures have been aligned (center of mass,
       orientation) previously.

       Arguments:
         | coor1  --  coordinates of structure 1 in a numpy array with shape (N,3)
         | coor2  --  coordinates of structure 2 in a numpy array with shape (N,3)

       Optional arguments:
         | masses
           --  when given, the mass-weighted delta vector is computed
         | normalize
           --  whether delta vector should be normalized [default=False]

    """
    # check consistency
    if len(coor1) != len(coor2):
        raise ValueError("coordinates should have same length: found %i and %i" % (len(coor1), len(coor2)))

    delta = np.ravel(coor1 - coor2)
    if not masses is None:  #Mass-weighting delta vector
        for i,mass in enumerate(masses):
            delta[3*i:3*(i+1)] *=  np.sqrt(mass)
    if normalize:   #Normalizing delta vector
        norm = np.sum(delta**2)
        delta /= np.sqrt(norm)
    return np.reshape(delta, (-1,1))


def compute_sensitivity_freq(nma, index, symmetric=False, massweight=True):
    """Compute the sensity of the index-th frequency to changes in the mass-weighted Hessian elements.

       Arguments:
         nma  --  an NMA object

       Optional arguments:
         | symmetric  --  when True, a slightly different formula is used to take
                          into account the symmetry of the Hessian [default=False]
         | massweight  --  when True, a mass-weighted hessian is considered
    """
    L = 3*len(nma.masses)
    mode = nma.modes[:,index]
    if not massweight: # un-mass-weight the mode
        for at,mass in enumerate(nma.masses):
            mode[3*at:3*(at+1)] /= np.sqrt(mass)
        mode = mode / np.sqrt(np.sum(mode**2))  # renormalization necessary

    mat = np.dot( np.reshape(mode,(L,1)), np.reshape(mode,(1,L)) )
    if symmetric:
        mat *= 2
        for i in range(L):
             mat[i,i] -= mode[i]**2
    return mat


def create_blocks_peptide_charmm(filename, label="normal", blocksize=1):
    """Create blocks list for CHARMM peptides

       Argument:
         | filename  --  the CHARMM coordinate file (typically extension .crd or
                         .cor)

       Optional argument:
         | label  --  type of MBH blocks: RTB, dihedral, RHbending, normal
                    [default=normal]
         | blocksize  --  when using the RTB scheme, blocksize defines the number
                        of residues in a block

       TODO: referenties
    """
    N, calpha, proline, carbon, oxygen, nitrogen = load_peptide_info_charmm(filename)

    if label is "RTB":
        return _calc_blocks_RTB(blocksize, N, calpha, proline, carbon, oxygen, nitrogen)
    elif label is "dihedral":
        return _calc_blocks_dihedral(N, calpha, proline, carbon, oxygen, nitrogen)
    elif label is "RHbending":
        return _calc_blocks_RHbending(N, calpha, proline, carbon, oxygen, nitrogen)
    elif label is "normal":
        return _calc_blocks_normal(N, calpha, proline, carbon, oxygen, nitrogen)
    else:
        raise NotImplementedError


def _get_pept_linked(N, calpha, proline):
    """Defines the peptide blocks [CA,C,O,N,NH,CA]

       The alpha carbons are shared between neighboring blocks. The routine
       relies on the conventional atom order in CHARMM peptide files.
    """
    # PEPT bonds = calpha + CONH + calpha = 6 atoms
    pept = []
    for i in xrange(1,len(calpha)):
        if calpha[i] in proline:
            pept.append( [calpha[i-1]] + range(calpha[i]-6,calpha[i]+1) )
            #print "proline found!", calpha[i]
        else:
            pept.append( [calpha[i-1]] + range(calpha[i]-4,calpha[i]+1) )
    return pept


def _calc_blocks_RTB(blocksize, N, calpha, proline, carbon, oxygen, nitrogen):
    """Rotation-Translation Blocks scheme of Tama et al

       Arguments: see functions 'create_blocks_peptide_charmm' and
                  'load_peptide_info_charmm'

       This amounts to applying the Mobile Block Hessian with
       one or more residues per block.
    """
    pept = []
    k = blocksize               # number of residues per block
    n = len(calpha)/blocksize   # number of complete RTB-blocks

    # do for first block : ... (calpha,R) x size - CO
    if calpha[k-1] > 1:
        if calpha[k-1] in proline:
            pept.append(range(1,calpha[k-1]-4))
        else:
            pept.append(range(1,calpha[k-1]-2))

    # for next blocks :  (N - calpha,R - CO) x size
    for i in xrange(1,n):   # do n-1 times
        # from first N till next N
        if calpha[i*k-1] in proline:
            if calpha[(i+1)*k-1] in proline:
                pept.append(range(calpha[i*k-1]-4,calpha[(i+1)*k-1]-4))
                #print "(side chain) proline found! (1,2)", calpha[i*k-1],calpha[(i+1)*k-1]
            else:
                pept.append(range(calpha[i*k-1]-4,calpha[(i+1)*k-1]-2))
                #print "(side chain) proline found! (1)", calpha[i*k-1]
        else:
            if calpha[(i+1)*k-1] in proline:
                pept.append(range(calpha[i*k-1]-2,calpha[(i+1)*k-1]-4))
                #print "(side chain) proline found! (2)", calpha[(i+1)*k-1]
            else:
                pept.append(range(calpha[i*k-1]-2,calpha[(i+1)*k-1]-2))

    # for last block : N - calpha,R
    if n*k-1 < len(calpha):
        if calpha[n*k-1] in proline:
            pept.append(range(calpha[n*k-1]-4,N+1))
            #print "(side chain) proline found! (1)", calpha[n*k-1]
        else:
            pept.append(range(calpha[n*k-1]-2,N+1))
    return pept


def _calc_blocks_dihedral(N, calpha, proline, carbon, oxygen, nitrogen):
    """MBH scheme with linked blocks that selects the characteristic Phi and Psi
       dihedral angles of the protein backbone as variables, while other
       degrees of freedom are fixed.

       Arguments: see function 'load_peptide_info_charmm'
    """
    pept = _get_pept_linked(N, calpha, proline)
    res  = []

    # start with an ending :   ... calpha_0,R - C
    if calpha[1] in proline:
        res.append( range(1,calpha[1]-5) )
        #print "proline found!", calpha[1]
    else:
        res.append( range(1,calpha[1]-3) )
    # continue with normal residues : N - calpha,R - C
    for i in xrange(1,len(calpha)-1):
        if calpha[i] in proline:
            if calpha[i+1] in proline:
                res.append( [calpha[i]-4]+range(calpha[i],calpha[i+1]-5) )
                #print "(side chain) proline found! (1,2)", calpha[i],calpha[i+1]
            else:
                res.append( [calpha[i]-4]+range(calpha[i],calpha[i+1]-3) )
                #print "(side chain) proline found! (1)", calpha[i]
        else:
            if calpha[i+1] in proline:
                res.append( [calpha[i]-2]+range(calpha[i],calpha[i+1]-5) )
                #print "(side chain) proline found! (2)", calpha[i+1]
            else:
                res.append( [calpha[i]-2]+range(calpha[i],calpha[i+1]-3) )
    # finish with another ending : N - calpha,R
    if calpha[-1] in proline:
        res.append([calpha[-1]-4]+range(calpha[-1],N+1))
        #print "(side chain) proline found!", calpha[-1]
    else:
        res.append([calpha[-1]-2]+range(calpha[-1],N+1))
    return res + pept


def _calc_blocks_RHbending(N, calpha, proline, carbon, oxygen, nitrogen):
    """MBH scheme in which the calpha-H bond is considered as a seperate block.

       Arguments: see function 'load_peptide_info_charmm'
    """
    pept = _get_pept_linked(N, calpha, proline)
    res  = []
    CH   = []

    # start with an ending
    if calpha[1] in proline:
        res.append( range(1,calpha[1]-6) )
        #print "(side chain) proline found!", calpha[1]
    else:
        res.append( range(1,calpha[1]-4) )
    # continue with normal residues
    for i in xrange(1,len(calpha)-1):  # calpha and HA
        CH.append( [calpha[i],calpha[i]+1] )
    for i in xrange(1,len(calpha)-1):  # calpha and rest of residue
        if calpha[i+1] in proline:
            res.append( [calpha[i]]+range(calpha[i]+2,calpha[i+1]-6) )
            #print "(side chain) proline found!", calpha[i+1]
        else:
            res.append( [calpha[i]]+range(calpha[i]+2,calpha[i+1]-4) )
    # finish with another ending
    CH.append([calpha[-1],calpha[-1]+1]) # calpha and HA
    res.append([calpha[-1]]+range(calpha[-1]+2,N+1)) # calpha and rest of residue

    return res + pept + CH


def _calc_blocks_normal(N, calpha, proline, carbon, oxygen, nitrogen):
    """MBH scheme with linked blocks where each side chain is considered as a
       block attached to the calpha atom.

       Arguments: see function 'load_peptide_info_charmm'
    """
    pept = _get_pept_linked(N, calpha, proline)
    res  = []

    # start with an ending
    if calpha[1] in proline:
        res.append( range(1,calpha[1]-6) )
        #print "(side chain) proline found!", calpha[1]
    else:
        res.append( range(1,calpha[1]-4) )
    # continue with normal residues
    for i in xrange(1,len(calpha)-1):
        if calpha[i+1] in proline:
            res.append( range(calpha[i],calpha[i+1]-6) )
            #print "(side chain) proline found!", calpha[i+1]
        else:
            res.append( range(calpha[i],calpha[i+1]-4) )
    # finish with another ending
    res.append(range(calpha[-1],N+1))

    return res + pept


def create_subs_peptide_charmm(filename, atomtypes=["CA"], frequency=1):
    """Create subsystem selection for CHARMM peptides

       Argument:
         | filename  --  the CHARMM coordinate file (typically extension .crd or
                       .cor)

       Optional argument:
         | atomtypes  --  list of strings. Let only these atom types be part of
                          the subsystem.
         | frequency  --  let only one out of every *frequency* residues be part
                          of the subsystem.
    """
    N, calpha, proline, carbon, oxygen, nitrogen = load_peptide_info_charmm(filename)

    subs = []
    if "C" in atomtypes:
        subs.extend(carbon[::frequency])
    if "O" in atomtypes:
        subs.extend(oxygen[::frequency])
    if "N" in atomtypes:
        subs.extend(nitrogen[::frequency])
    if "CA" in atomtypes:
        subs.extend(calpha[::frequency])
    if len(subs) == 0:
        raise NotImplementedError("No matching atom types found.")
    return subs


def plot_spectrum_lines(filename, all_freqs, low=None, high=None, title=None):
    """Plot multiple spectra in a comparative line plot

       Arguments:
         | filename  --  the filename to write the figure to (the extension and
                         the matplotlib settings determine the file format)
         | all_freqs  --  a list with spectra, each item in the list is an array
                         with multiple frequencies that represent one spectrum

       Optional arguments:
         | low  --  minimum on x-axis, in atomic units
         | high  --  maximum on x-axis, in atomic units
         | title  --  title for plot (a string)
    """
    import matplotlib.pyplot as pt
    pt.clf()
    for i, freqs in enumerate(all_freqs):
        for freq in freqs:
            if (freq>low or low is None) and (freq<high or high is None):
                pt.plot([i+0.75,i+1.25],[freq/invcm,freq/invcm],"k-")
    pt.xticks(range(1,len(all_freqs)+1))
    if low is not None:
        pt.ylim(ymin=low/invcm)
    if high is not None:
        pt.ylim(ymax=high/invcm)
    pt.ylabel("Frequency in cm-1")
    if title is not None:
        pt.title(title)
    pt.savefig(filename)


def plot_spectrum_dos(filename, all_freqs, low=None, high=None, imax=None,
                      step=1.0*invcm, width=10.0*invcm, all_amps=None, title=None):
    """Plot multiple spectra in a comparative density of states plot

       Arguments:
         | filename  --  the filename to write the figure too (the extension and
                         the matplotlib settings determine the file format)
         | all_freqs  --  a list with spectra, each item in the list is an array
                         with multiple frequencies that represent one spectrum

       Optional arguments:
         | low  --  minimum on x-axis, in atomic units
         | high  --  maximum on x-axis, in atomic units
         | imax  --  maximum intensity on y-axis, no unit
         | step  --  resulotion of plot, in atomic units
         | width  --  width of Gaussian, in atomic units
         | all_amps  --  list of arrays in the same format as all_freqs with an
                         amplitude for each individual frequency
         | title  --  title for plot (a string)
    """
    import matplotlib.pyplot as pt

    def plot_single_dos(freqs, low, high, step, width, amps):
        if low is None:
            low=freqs[0]-3*width   # to avoid unnecessary plotting
        if high is None:
            high=freqs[-1]+3*width

        nu = np.arange(low, high, step)
        intensity = np.zeros(nu.shape, float)

        s2 = width**2 / ( 8*np.log(2) )  # standard deviation squared
        for i in xrange(len(freqs)):
            if freqs[i] > low and freqs[i] < high:
                tmp = np.exp(-(nu-freqs[i])**2/(2*s2))
                if amps is not None:
                    if hasattr(amps, "__len__"):
                        tmp *= amps[i]
                    else:
                        tmp *= amps
                intensity += tmp
        pt.plot(nu/invcm, intensity)

    pt.clf()
    for i, freqs in enumerate(all_freqs):
        if all_amps is None:
            amps = None
        else:
            amps = all_amps[i]
        plot_single_dos(freqs, low, high, step, width, amps)
    pt.ylim(0.0, imax)
    if low is not None:
        pt.xlim(xmin=low/invcm)
    if high is not None:
        pt.xlim(xmax=high/invcm)
    pt.legend([str(nb) for nb in range(1,len(all_freqs)+1)])
    pt.xlabel("Frequency in cm-1")
    pt.ylabel("Intensity")
    if title is not None:
        pt.title(title)
    pt.savefig(filename)


def create_enm_molecule(molecule, selected=None, numbers=None, masses=None,
                        rcut=8.0*angstrom, K=1.0, periodic=None):
    """Create a molecule according to the Elastic Network Model

       Argument:
         | molecule  --  The molecule to start from. can be two types: (i) a
                       Molecule object or (ii) a numpy array with shape (N,3)
                       with coordinates in atomic units.

       When a Molecule object is given, atom numbers, masses and periodic are
       inherited from the molecule, unless they are specified explicitly in the
       optional arguments.

       Optional arguments:
         | selected  --  Selection of atoms to include in the ENM model. This can
                         be a list or array of atom indices (length <= N), or an
                         array of booleans (length = N).
         | numbers  --  atom numbers in the ENM model (length = N). default is
                        array of ones or the numbers from the molecule object.
         | masses  --  atomic masses in atomic units in the ENM model (length = N).
                       default is array of hydrogen masses or the masses from the
                       molecule object.
         | rcut  --  cutoff distance between interacting pairs in atomic units
         | K  --  strength of the interaction in atomic units (Hartree/Bohr**2).
                  The interaction strength is the same for all interacting pairs.
    """
    if isinstance(molecule, Molecule):
        coordinates = molecule.coordinates
        if numbers is None:
            numbers = molecule.numbers
        if masses is None:
            masses = molecule.masses
        if periodic is None:
            periodic = molecule.periodic
    else:
        coordinates = np.array(molecule, copy=False)
        if numbers is None:
            # pretend being hydrogens
            numbers = np.ones(len(coordinates))
        if masses is None:
            # pretend being hydrogens
            masses  = np.ones(len(coordinates),float)*amu
        if periodic is None:
            periodic = False

    if selected is not None:
        coordinates = coordinates[selected]
        numbers = numbers[selected]
        masses = masses[selected]

    rcut2 = rcut**2
    N = len(coordinates)
    hessian = np.zeros((3*N,3*N),float)
    for i in range(N):
        for j in range(i+1,N):
            x = np.reshape( (coordinates[i,:]-coordinates[j,:]), (3,1))
            dist2 = np.sum(x**2)
            if dist2 < rcut2:
                corr = K * np.dot( x, x.transpose() ) / dist2
                hessian[3*i:3*(i+1), 3*i:3*(i+1)] += corr
                hessian[3*i:3*(i+1), 3*j:3*(j+1)] -= corr
                hessian[3*j:3*(j+1), 3*i:3*(i+1)] -= corr
                hessian[3*j:3*(j+1), 3*j:3*(j+1)] += corr

    return Molecule(
        numbers,
        coordinates,
        masses,
        0.0, #energy
        np.zeros((len(coordinates),3), float),  #gradient
        hessian,
        1, # multiplicity
        1, # rotational symmetry number
        periodic,
    )



# -*- coding: utf-8 -*-
# TAMkin is a post-processing toolkit for normal mode analysis, thermochemistry
# and reaction kinetics.
# Copyright (C) 2008-2012 Toon Verstraelen <Toon.Verstraelen@UGent.be>, An Ghysels
# <An.Ghysels@UGent.be> and Matthias Vandichel <Matthias.Vandichel@UGent.be>
# Center for Molecular Modeling (CMM), Ghent University, Ghent, Belgium; all
# rights reserved unless otherwise stated.
#
# This file is part of TAMkin.
#
# TAMkin is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# In addition to the regulations of the GNU General Public License,
# publications and communications based in parts on this program or on
# parts of this program are required to cite the following article:
#
# "TAMkin: A Versatile Package for Vibrational Analysis and Chemical Kinetics",
# An Ghysels, Toon Verstraelen, Karen Hemelsoet, Michel Waroquier and Veronique
# Van Speybroeck, Journal of Chemical Information and Modeling, 2010, 50,
# 1736-1750W
# http://dx.doi.org/10.1021/ci100099g
#
# TAMkin is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --


import os

import numpy as np
import pkg_resources

from molmod.periodic import periodic
from molmod import angstrom, amu, calorie, avogadro, electronvolt, lightspeed, \
    UnitCell
from molmod.test.common import tmpdir

from tamkin import *


def test_get_submolecule():
    molecule = load_molecule_charmm(
        pkg_resources.resource_filename(__name__, "../data/test/an/ethanol.cor"),
        pkg_resources.resource_filename(__name__, "../data/test/an/ethanol.hess.full"))
    select = [3,4,2]
    molecule2 = molecule.get_submolecule(select)
    for i,at in enumerate(select):
        assert molecule.numbers[at] == molecule2.numbers[i]
        assert abs(molecule.masses[at] - molecule2.masses[i]) < 1e-3
        assert molecule.symbols[at] == molecule2.symbols[i]
        for mu in range(3):
            assert abs(molecule.coordinates[at,mu] - molecule2.coordinates[i,mu]) < 1e-5
            assert abs(molecule.gradient[at,mu] - molecule2.gradient[i,mu]) < 1e-5
            for j,at2 in enumerate(select):
                for nu in range(3):
                    assert abs(molecule.hessian[3*at+mu,3*at2+nu] - molecule2.hessian[3*i+mu,3*j+nu]) < 1e-5
    assert molecule.multiplicity == molecule2.multiplicity
    assert molecule.symmetry_number == molecule2.symmetry_number
    assert molecule.periodic == molecule2.periodic
    assert molecule.energy == molecule2.energy


def test_get_submolecule_cp2k():
    molecule = load_molecule_cp2k(
        pkg_resources.resource_filename(__name__, "../data/test/cp2k/pentane/sp.out"),
        pkg_resources.resource_filename(__name__, "../data/test/cp2k/pentane/freq.out"))
    select = range(5)+[9,11,14]
    molecule2 = molecule.get_submolecule(select, title="this is submol", energy=5., periodic=False, symmetry_number=6)  # just trying out something
    for i,at in enumerate(select):
        assert molecule.numbers[at] == molecule2.numbers[i]
        assert abs(molecule.masses[at] - molecule2.masses[i]) < 1e-3
        for mu in range(3):
            assert abs(molecule.coordinates[at,mu] - molecule2.coordinates[i,mu]) < 1e-5
            assert abs(molecule.gradient[at,mu] - molecule2.gradient[i,mu]) < 1e-5
            for j,at2 in enumerate(select):
                for nu in range(3):
                    assert abs(molecule.hessian[3*at+mu,3*at2+nu] - molecule2.hessian[3*i+mu,3*j+nu]) < 1e-5
    assert molecule.multiplicity == molecule2.multiplicity
    assert molecule2.symmetry_number == 6
    assert molecule2.periodic == False
    assert molecule2.energy == 5.


def test_translate_pbc():
    molecule = load_molecule_cp2k(
        pkg_resources.resource_filename(__name__, "../data/test/cp2k/pentane/sp.out"),
        pkg_resources.resource_filename(__name__, "../data/test/cp2k/pentane/freq.out"))
    assert abs(molecule.unit_cell.matrix[1,1]/angstrom - 30.000) < 1e-3
    assert abs(molecule.unit_cell.matrix[0,1]/angstrom - 0.000) < 1e-3
    assert abs(molecule.coordinates[5,1]/angstrom - 13.928520) < 1e-5
    selected = range(6)+[11,14]
    molecule2 = translate_pbc(molecule, selected, [1,-1,0])
    assert abs(molecule2.coordinates[5,1]/angstrom -  (13.928520-30.0)) < 1e-5
    for i in range(molecule.size):
        assert molecule.numbers[i] == molecule2.numbers[i]
        assert abs(molecule.masses[i] - molecule2.masses[i]) < 1e-3
        for mu in range(3):
            if i not in selected:
                assert abs(molecule.coordinates[i,mu] - molecule2.coordinates[i,mu]) < 1e-3
            else:
                if mu is 0:
                    assert abs(molecule.coordinates[i,mu] + 30.*angstrom - molecule2.coordinates[i,mu]) < 1e-3
                if mu is 1:
                    assert abs(molecule.coordinates[i,mu] - 30.*angstrom - molecule2.coordinates[i,mu]) < 1e-3
                if mu is 2:
                    assert abs(molecule.coordinates[i,mu] - molecule2.coordinates[i,mu]) < 1e-3
            assert abs(molecule.gradient[i,mu] - molecule2.gradient[i,mu]) < 1e-3
            for j in selected:
                for nu in range(3):
                    assert abs(molecule.hessian[3*i+mu,3*j+nu] - molecule2.hessian[3*i+mu,3*j+nu]) < 1e-3
    assert molecule.multiplicity == molecule2.multiplicity
    assert molecule.symmetry_number == molecule2.symmetry_number
    assert abs(molecule2.unit_cell.matrix[0,0]/angstrom - 30.000) < 1e-3
    assert abs(molecule2.unit_cell.matrix[1,2]/angstrom - 0.000) < 1e-3


def test_molecule_checkpoint_basic():
    mol1 = load_molecule_g03fchk(
        pkg_resources.resource_filename(__name__, "../data/test/sterck/aa.fchk"))
    with tmpdir(__name__, 'test_molecule_checkpoint_basic') as dn:
        fn_out = os.path.join(dn, "molecule_checkpoint_basic.chk")
        mol1.write_to_file(fn_out)
        mol2 = Molecule.read_from_file(fn_out)

    assert mol1.numbers.shape == mol2.numbers.shape
    assert abs(mol1.numbers - mol2.numbers).max() == 0
    assert mol1.coordinates.shape == mol2.coordinates.shape
    assert abs(mol1.coordinates - mol2.coordinates).max() < 1e-10
    assert mol1.masses.shape == mol2.masses.shape
    assert abs(mol1.masses - mol2.masses).max() < 1e-10
    assert abs(mol1.energy - mol2.energy) < 1e-5
    assert mol1.gradient.shape == mol2.gradient.shape
    assert abs(mol1.gradient - mol2.gradient).max() < 1e-10
    assert mol1.hessian.shape == mol2.hessian.shape
    assert abs(mol1.hessian - mol2.hessian).max() < 1e-10
    assert mol1.multiplicity == mol2.multiplicity
    assert mol1.symmetry_number == mol2.symmetry_number
    assert mol1.periodic == mol2.periodic


def test_molecule_checkpoint_full():
    mol1 = load_molecule_g03fchk(
        pkg_resources.resource_filename(__name__, "../data/test/sterck/aa.fchk"))
    mol1.set_default_graph()
    mol1.unit_cell = UnitCell(np.identity(3, float)*25)
    mol1.symbols = [periodic[n].symbol for n in mol1.numbers]
    with tmpdir(__name__, 'test_molecule_checkpoint_full') as dn:
        fn_out = os.path.join(dn, "molecule_checkpoint_full.chk")
        mol1.write_to_file(fn_out)
        mol2 = Molecule.read_from_file(fn_out)

    assert mol1.numbers.shape == mol2.numbers.shape
    assert abs(mol1.numbers - mol2.numbers).max() == 0
    assert mol1.coordinates.shape == mol2.coordinates.shape
    assert abs(mol1.coordinates - mol2.coordinates).max() < 1e-10
    assert mol1.masses.shape == mol2.masses.shape
    assert abs(mol1.masses - mol2.masses).max() < 1e-10
    assert abs(mol1.energy - mol2.energy) < 1e-10
    assert mol1.gradient.shape == mol2.gradient.shape
    assert abs(mol1.gradient - mol2.gradient).max() < 1e-10
    assert mol1.hessian.shape == mol2.hessian.shape
    assert abs(mol1.hessian - mol2.hessian).max() < 1e-10
    assert mol1.multiplicity == mol2.multiplicity
    assert mol1.symmetry_number == mol2.symmetry_number
    assert mol1.periodic == mol2.periodic

    assert mol1.graph.edges == mol2.graph.edges
    assert mol1.title == mol2.title
    assert mol1.unit_cell.matrix.shape == mol2.unit_cell.matrix.shape
    assert abs(mol1.unit_cell.matrix - mol2.unit_cell.matrix).max() < 1e-10
    assert mol1.unit_cell.active.shape == mol2.unit_cell.active.shape
    assert (mol1.unit_cell.active == mol2.unit_cell.active).all()
    assert mol1.symbols == mol2.symbols


def test_copy_with():
    mol1 = load_molecule_g03fchk(
        pkg_resources.resource_filename(__name__, "../data/test/sterck/aa.fchk"))
    mol2 = mol1.copy_with(title="foo")
    assert mol2.title == "foo"


def test_get_external_basis_new1():
    mol = load_molecule_g03fchk(
        pkg_resources.resource_filename(__name__, "../data/test/linear/gaussian.fchk"))
    eb = mol.get_external_basis_new()
    assert eb.shape[0] == 5
    # orthogonality test: the external basis should be orthogonal
    # to the basis of the four internal coordinates: stretch1, stertch2,
    # bend1 and bend2
    ib = np.array([
        [0,0,1,0,0,-1,0,0,0],
        [0,0,1,0,0,0,0,0,-1],
        [0,-1,0,0,0.5,0,0,0.5,0],
        [-1,0,0,0.5,0,0,0.5,0,0],
    ])
    error = abs(np.dot(ib, eb.transpose())).max()
    assert error < 1e-5


def test_get_external_basis_new2():
    mol = load_molecule_g03fchk(
        pkg_resources.resource_filename(__name__, "../data/test/ethane/gaussian.fchk"))
    eb = mol.get_external_basis_new()
    assert eb.shape[0] == 6
    # orthogonality test: the external basis should be orthogonal
    # to the basis of the internal coordinates. in this case we construct
    # a redundant basis of internal coordinates based on all interatomic
    # distances. this is overkill...
    ib = []
    for i in xrange(mol.size):
        for j in xrange(i):
            b = np.zeros((mol.size, 3), float)
            delta = mol.coordinates[j] - mol.coordinates[i]
            delta /= np.linalg.norm(delta)
            b[i] = delta
            b[j] = -delta
            ib.append(b.ravel())
    ib = np.array(ib)
    error = abs(np.dot(ib, eb.transpose())).max()
    assert error < 1e-5


def test_rot_scan_ts():
    # The select dihedral angles do not allow an automatic assignment of the top
    # indexes.
    mol = load_molecule_g03fchk(
        pkg_resources.resource_filename(__name__, "../data/test/sterck/paats_1h2o_b_aa.fchk"))
    diheds = [[9, 6, 7, 10], [7, 10, 18, 11], [11, 18, 10, 7]]
    for dihed in diheds:
        try:
            scan = RotScan(np.array(dihed)-1, mol)
            assert False
        except ValueError, e:
            assert e.message == "The rotating top could not be assigned properly. Specify the top_indexes manually."



# -*- coding: utf-8 -*-
# TAMkin is a post-processing toolkit for normal mode analysis, thermochemistry
# and reaction kinetics.
# Copyright (C) 2008-2012 Toon Verstraelen <Toon.Verstraelen@UGent.be>, An Ghysels
# <An.Ghysels@UGent.be> and Matthias Vandichel <Matthias.Vandichel@UGent.be>
# Center for Molecular Modeling (CMM), Ghent University, Ghent, Belgium; all
# rights reserved unless otherwise stated.
#
# This file is part of TAMkin.
#
# TAMkin is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# In addition to the regulations of the GNU General Public License,
# publications and communications based in parts on this program or on
# parts of this program are required to cite the following article:
#
# "TAMkin: A Versatile Package for Vibrational Analysis and Chemical Kinetics",
# An Ghysels, Toon Verstraelen, Karen Hemelsoet, Michel Waroquier and Veronique
# Van Speybroeck, Journal of Chemical Information and Modeling, 2010, 50,
# 1736-1750W
# http://dx.doi.org/10.1021/ci100099g
#
# TAMkin is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
#--


from tamkin import *

corfile = "open/adk.open.new.cor"
fixedfile = "fixed_files/fixed."


for i in range(1,6):
    blocks = create_blocks_peptide_charmm(corfile, BlocksPeptideMBH("RTB", blocksize=i))
    blocks_write_to_file(blocks, fixedfile+str(i)+".txt")

for i in range(1,6):
    subs = create_subs_peptide_charmm(corfile, SubsPeptideVSA(frequency=i))
    subs_write_to_file(subs, fixedfile+str(i+5)+".txt")


blocks = create_blocks_peptide_charmm(corfile, BlocksPeptideMBH("RHbending"))
blocks_write_to_file(blocks, fixedfile+"11.txt")

blocks = create_blocks_peptide_charmm(corfile, BlocksPeptideMBH("dihedral"))
blocks_write_to_file(blocks, fixedfile+"12.txt")

blocks = create_blocks_peptide_charmm(corfile, BlocksPeptideMBH())
blocks_write_to_file(blocks, fixedfile+"13.txt")


#for x in 6 7 8 9 10 ; do
#python analyse-adk.py --job vsa --filecor open/adk.open.new.cor  --filehess open/adk.open.hess.full \
#                      --filechk chkfiles/adk.open.vsa.$x.chk  \
#                      --filefixed fixed_files/fixed.$x.txt

#done



# -*- coding: utf-8 -*-
# TAMkin is a post-processing toolkit for normal mode analysis, thermochemistry
# and reaction kinetics.
# Copyright (C) 2008-2012 Toon Verstraelen <Toon.Verstraelen@UGent.be>, An Ghysels
# <An.Ghysels@UGent.be> and Matthias Vandichel <Matthias.Vandichel@UGent.be>
# Center for Molecular Modeling (CMM), Ghent University, Ghent, Belgium; all
# rights reserved unless otherwise stated.
#
# This file is part of TAMkin.
#
# TAMkin is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# In addition to the regulations of the GNU General Public License,
# publications and communications based in parts on this program or on
# parts of this program are required to cite the following article:
#
# "TAMkin: A Versatile Package for Vibrational Analysis and Chemical Kinetics",
# An Ghysels, Toon Verstraelen, Karen Hemelsoet, Michel Waroquier and Veronique
# Van Speybroeck, Journal of Chemical Information and Modeling, 2010, 50,
# 1736-1750W
# http://dx.doi.org/10.1021/ci100099g
#
# TAMkin is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --


import os
import unittest

from molmod.test.common import tmpdir

from tamkin import *


__all__ = ["TimerTestCase"]


class TimerTestCase(unittest.TestCase):
    def test_timer(self):
        timer = Timer()
        timer.sample("start")
        for i in range(100000):
            j = i**2
        timer.sample("done")
        self.assertEqual(timer.labels[0], "start")
        self.assertEqual(timer.labels[1], "done")
        #timer.dump()  # to write logfile to screen
        # just testing whether writing to file works
        with tmpdir(__name__, 'test_timer') as dn:
            timer.write_to_file(os.path.join(dn, "logfile-timings.txt"))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# TAMkin is a post-processing toolkit for normal mode analysis, thermochemistry
# and reaction kinetics.
# Copyright (C) 2008-2012 Toon Verstraelen <Toon.Verstraelen@UGent.be>, An Ghysels
# <An.Ghysels@UGent.be> and Matthias Vandichel <Matthias.Vandichel@UGent.be>
# Center for Molecular Modeling (CMM), Ghent University, Ghent, Belgium; all
# rights reserved unless otherwise stated.
#
# This file is part of TAMkin.
#
# TAMkin is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# In addition to the regulations of the GNU General Public License,
# publications and communications based in parts on this program or on
# parts of this program are required to cite the following article:
#
# "TAMkin: A Versatile Package for Vibrational Analysis and Chemical Kinetics",
# An Ghysels, Toon Verstraelen, Karen Hemelsoet, Michel Waroquier and Veronique
# Van Speybroeck, Journal of Chemical Information and Modeling, 2010, 50,
# 1736-1750W
# http://dx.doi.org/10.1021/ci100099g
#
# TAMkin is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
#--


from lot_basis import lots_list
import html
from kin import *

from molmod.units import kjmol
from molmod.constants import boltzmann

import matplotlib.pyplot as pt, numpy

from matplotlib.backend_bases import GraphicsContextBase
GraphicsContextBase.dashd["dashed"] = (0, (6.0, 3.0))
GraphicsContextBase.dashd["dashdot"] = (0, (4.0, 2.0, 1.0, 2.0))
GraphicsContextBase.dashd["dotted"] = (0, (1.5, 1.5))




def overview(template, title, fn_img, rows):
    if len(rows) == 0:
        rows.append([u"<th>Approach→<br />Functional↓</th>"])
        rows.append(["<th></th>"])
    rows[0].append("<th colspan=4>%s</th>" % title.replace(", ", "<br />"))
    for temp in temps:
        rows[1].append("<td style='font-size:small'>%.0fK</td>" % temp)
    lines = []
    labels = []
    pt.clf()
    line = pt.plot(invtemps, experimental_k, color="k", linestyle="-", lw=4)
    lines.append(line)
    labels.append("experiment")
    counter = 2
    for lot in lots_list:
        if lot.spin == "ROS":
            lot_label = "ro" + lot.label
        else:
            lot_label = lot.label
        if len(rows) <= counter:
            rows.append(["<th>%s</th>" % lot_label])
        try:
            ks = load_summary(template % lot_label)[0]
            line = pt.plot(invtemps, ks, color=lot.color, linestyle=lot.linestyle, lw=2)
            lines.append(line)
            labels.append(lot_label)
            for j in xrange(4):
                ln10ratio = numpy.log10(ks[j]/experimental_k[j])
                color = get_error_color(ln10ratio)
                rows[counter].append("<td style='background-color:%s'>%.0f</td>" % (color, ln10ratio*10))
        except IOError, StopIteration:
            rows[counter].append("<td>&nbsp</td><td>&nbsp</td><td>&nbsp</td><td>&nbsp</td>")
        counter += 1
    pt.semilogy()
    pt.fill(
        numpy.concatenate((invtemps, invtemps[::-1])),
        numpy.concatenate((experimental_k*10, experimental_k[::-1]/10)),
        "k", alpha=0.2, lw=0,
    )
    pt.xticks(
        1/numpy.array([300, 350, 400, 450, 500, 550, 600], float),
        ["300", "350", "400", "450", "500", "550", "600"],
    )
    pt.title(title)
    pt.xlabel("T [K]")
    pt.ylabel("k [(m**3/mol)/s]")
    pt.ylim(1e-8,1e7)
    legend = pt.figlegend(
        lines, labels, (0.07,0.06), ncol=3, prop={"size":11},
        handlelength=3, labelspacing=0.1, columnspacing=1
    )
    #legend.get_frame().set_linewidth(0)
    legend.get_frame().set_fill(True)
    legend.get_frame().set_alpha(0.5)
    pt.savefig(fn_img % "rates")

    pt.clf()
    lines = []
    labels = []
    line = pt.plot([experimental_Ea], [experimental_A], color="k", marker="o", ms=11, mew=2, lw=0, ls=" ")
    lines.append(line)
    labels.append("experiment")
    for lot in lots_list:
        if lot.spin == "ROS":
            label = "ro" + lot.label
        else:
            label = lot.label
        try:
            A, Ea = load_summary(template % label)[1]
            marker = {"-": "o", "--": "s", ":": "v", "-.": "h"}[lot.linestyle]
            line = pt.plot([Ea], [A], color=lot.color, marker=marker, ms=11, mew=2, lw=0, ls=" ")
            lines.append(line)
            labels.append(label)
            continue
        except IOError, StopIteration:
            pass

    pt.title(title)
    pt.xlabel("Activation energy [kJ/mol]")
    pt.ylabel("Pre-exponential factor [(m**3/mol)/s]")
    pt.semilogy()
    # error margin around experimental data point
    x = []
    y = []
    evals, evecs = numpy.linalg.eigh(covariance_parameters)
    angles = numpy.arange(0.0,360.5,1.0)/180*numpy.pi
    data = numpy.outer(evecs[:,0],numpy.cos(angles))*numpy.sqrt(evals[0]) + \
           numpy.outer(evecs[:,1],numpy.sin(angles))*numpy.sqrt(evals[1])
    pt.fill(
        experimental_parameters[1] + data[1],
        numpy.exp(experimental_parameters[0] + data[0]),
        color="k", alpha=0.2, lw=0
    )
    # end error margin
    legend = pt.legend(
        lines, labels, loc=4, ncol=4, prop={"size":11},
        handlelength=1, labelspacing=0.2, columnspacing=2,
        numpoints=1,
    )
    pt.xlim(0,90)
    pt.ylim(1e3,1e7)
    #legend.get_frame().set_linewidth(0)
    legend.get_frame().set_fill(True)
    legend.get_frame().set_alpha(0.5)
    pt.savefig(fn_img % "params")




f = file("kintab.html", "w")
print >> f, html.header % "KIN Overview"

for do_rotor in False, True:
    ir_str = {True: "ir", False: "ho"}[do_rotor]
    ir_info = {
        True: "internal rotor",
        False: "harmonic oscillator",
    }[do_rotor]
    for do_counterpoise in False, True:
        cp_str = {True: "cps", False: "bss"}[do_counterpoise]
        cp_info = {
            True: "with counterpoise correction",
            False: "without counterpoise correction",
        }[do_counterpoise]

        rows = []

        for ts_conformer in "Gauche", "Trans":
            overview(
                "%%s__6-31gd/%s_%s_summary_%s.txt" % (ir_str, cp_str, ts_conformer.lower()),
                "%s, %s, %s, Consistent, 6-31G(d)" % (ir_str.upper(), cp_str.upper(), ts_conformer),
                "kin_%s_%s_%s_consistent_6-31gd_%%s.pdf" % (ir_str, cp_str, ts_conformer.lower()),
                rows,
            )
        for ts_conformer in "Gauche", "Trans":
            overview(
                "%%s__6-311+g3df2p/%s_%s_summary_%s.txt" % (ir_str, cp_str, ts_conformer.lower()),
                "%s, %s, %s, Consistent, 6-311+G(3df,2p)" % (ir_str.upper(), cp_str.upper(), ts_conformer),
                "kin_%s_%s_%s_consistent_6-311+g3df2p_%%s.pdf" % (ir_str, cp_str, ts_conformer.lower()),
                rows,
            )
        for ts_conformer in "Gauche", "Trans":
            overview(
                "GEO__b3lyp__6-31gd__ENERGY__%%s__6-311+g3df2p/%s_%s_summary_%s.txt" % (ir_str, cp_str, ts_conformer.lower()),
                "%s, %s, %s, GEO=B3LYP/6-31G(d), 6-311+G(3df,2p)" % (ir_str.upper(), cp_str.upper(), ts_conformer),
                "kin_%s_%s_%s_geo_6-311+g3df2p_%%s.pdf" % (ir_str, cp_str, ts_conformer.lower()),
                rows,
            )

        print >> f, "<p>10Log10 of ratio between theoretical and experimental rate (%s, %s)</p>" % (ir_info, cp_info)
        html.print_table(f, rows)

print >> f, html.footer



# -*- coding: utf-8 -*-
# TAMkin is a post-processing toolkit for normal mode analysis, thermochemistry
# and reaction kinetics.
# Copyright (C) 2008-2012 Toon Verstraelen <Toon.Verstraelen@UGent.be>, An Ghysels
# <An.Ghysels@UGent.be> and Matthias Vandichel <Matthias.Vandichel@UGent.be>
# Center for Molecular Modeling (CMM), Ghent University, Ghent, Belgium; all
# rights reserved unless otherwise stated.
#
# This file is part of TAMkin.
#
# TAMkin is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# In addition to the regulations of the GNU General Public License,
# publications and communications based in parts on this program or on
# parts of this program are required to cite the following article:
#
# "TAMkin: A Versatile Package for Vibrational Analysis and Chemical Kinetics",
# An Ghysels, Toon Verstraelen, Karen Hemelsoet, Michel Waroquier and Veronique
# Van Speybroeck, Journal of Chemical Information and Modeling, 2010, 50,
# 1736-1750W
# http://dx.doi.org/10.1021/ci100099g
#
# TAMkin is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
"""Convenient interfaces to define thermodynamic and kinetic models"""


import numpy as np, csv

from molmod import boltzmann, kjmol, second, meter, mol, planck

from tamkin.partf import PartFun


__all__ = [
    "BaseModel", "ThermodynamicModel", "BaseKineticModel",
    "KineticModel", "ActivationKineticModel"
]


class BaseModel(object):
    """Base class for all physico-chemical models."""
    def __init__(self):
        """
           Useful attribute:
            | ``pfs_all`` -- A dictionary with all partition functions involved
                             in the model. The values are the partition function
                             objects. The keys are the corresponding signed
                             stoichiometries.

           The methods in the base class are mainly used by the Monte Carlo
           routine in the ReactionAnalysis class.
        """
        self.pfs_all = {}
        self.pfs_list = []

    def _add_pfs(self, pfs, factor=1.0):
        """Add partition functions to the global list.

           Argument:
            | ``pfs`` -- A list of partition functions.

           Optional argument
            | ``factor`` -- The stoichiometry of all partition functions will be
                            multiplied with this factor.
        """
        for pf in pfs:
            if isinstance(pf, PartFun):
                add_st = 1.0
            else:
                # It may also be a tuple of a partition function and a
                # stoichiometry.
                pf, add_st = pf
            st = self.pfs_all.get(pf)
            if st is None:
                # This is used to maintain an ordered list of partition functions
                self.pfs_list.append(pf)
                st = 0.0
            self.pfs_all[pf] = st + factor*add_st

    def _set_unit(self, per_second=False):
        """Set the unit of the equilibrium constant or pre-exponential factor.

           Optional argument:
            | ``per_second`` -- Boolean, when True the unit is divided by
                                second.

           This functions is called by the constructor after all partition
           functions are added. (See :meth:`_add_pfs`.)
        """
        meter_power = 0
        mol_power = 0
        for pf, st in self._iter_pfs():
            if hasattr(pf, "translational"):
                meter_power -= pf.translational.dim*st
            mol_power += st
        unit = mol**mol_power*meter**meter_power
        unit_name = ""
        if meter_power != 0:
            if meter_power == 1:
                unit_name += "m"
            else:
                unit_name += "m^%i" % meter_power
        if mol_power != 0:
            if mol_power == 1:
                unit_name += " mol"
            else:
                unit_name += " mol^%i" % mol_power
        if per_second:
            unit /= second
            if len(unit_name) != 0:
                unit_name += " "
            unit_name += "s^-1"
        self.unit = unit
        self.unit_name = unit_name

    def _iter_pfs(self):
        """Iterate over all partition functions with their stoichiometries."""
        for pf in self.pfs_list:
            st = self.pfs_all[pf]
            if abs(st) > 0:
                yield pf, st

    def backup_freqs(self):
        """Keep a backup copy of the frequencies and the energy of each partition function."""
        for pf in self.pfs_all:
            pf.vibrational.positive_freqs_orig = pf.vibrational.positive_freqs.copy()
            pf.vibrational.negative_freqs_orig = pf.vibrational.negative_freqs.copy()
            pf.electronic.energy_backup = pf.electronic.energy

    def alter_freqs(self, freq_error, scale_energy):
        """Randomly distort the frequencies and energies.

           Arguments:
            | ``freq_error`` -- The absolute error to be introduced in the
                                frequencies.
            | ``scale_energy`` -- The relative error to be introduced in the
                                  (electronic) energies.
        """
        for pf in self.pfs_all:
            N = len(pf.vibrational.positive_freqs)
            freq_shift = np.random.normal(0, freq_error, N)
            pf.vibrational.positive_freqs = pf.vibrational.positive_freqs_orig + freq_shift
            pf.vibrational.positive_freqs[pf.vibrational.positive_freqs<=0] = 0.01
            N = len(pf.vibrational.negative_freqs)
            freq_shift = np.random.normal(0, freq_error, N)
            pf.vibrational.negative_freqs = pf.vibrational.negative_freqs_orig + freq_shift
            pf.vibrational.negative_freqs[pf.vibrational.negative_freqs>=0] = -0.01
            pf.electronic.energy = pf.electronic.energy_backup*scale_energy

    def restore_freqs(self):
        """Restore the backup of the frequencies and the energy of each partition function."""
        for pf in self.pfs_all:
            pf.vibrational.positive_freqs = pf.vibrational.positive_freqs_orig
            pf.vibrational.negative_freqs = pf.vibrational.negative_freqs_orig
            pf.electronic.energy = pf.electronic.energy_backup
            del pf.vibrational.positive_freqs_orig
            del pf.vibrational.negative_freqs_orig
            del pf.electronic.energy_backup

    def free_energy_change(self, temp):
        """Compute the change in free energy.

           The change in free energy depends on the the pressure in the
           ``ExtTrans`` contribution to the partition function, if such a
           contribution would be present.

           Argument:
            | ``temp``  -- the temperature
        """
        return sum(pf.chemical_potential(temp)*st for pf, st in self._iter_pfs())

    def energy_difference(self):
        """Compute the electronic energy difference between (+) products and (-) reactants."""
        return sum(pf.electronic.energy*st for pf, st in self._iter_pfs())

    def zero_point_energy_difference(self):
        """Compute the zero-point energy difference between (+) products and (-) reactants."""
        return sum(pf.zero_point_energy()*st for pf, st in self._iter_pfs())

    def internal_heat_difference(self, temp):
        """Compute the internal_heat difference between (+) products and (-) reactants.

           Argument:
            | ``temp`` -- The temperature.
        """
        return sum(pf.internal_heat(temp)*st for pf, st in self._iter_pfs())

    def equilibrium_constant(self, temp, do_log=False):
        """Compute the equilibrium constant at the given temperature.

           Argument:
            | ``temp`` -- The temperature.

           Optional argument:
            | ``do_log`` -- When True, the logarithm of the equilibrium constant
                            is returned instead of just the equilibrium constant
                            itself. [default=False]
        """
        log_K = sum(pf.logv(temp)*st for pf, st in self._iter_pfs())
        if do_log:
            return log_K
        else:
            return np.exp(log_K)

    def write_table(self, temp, filename):
        """Write a CSV file with the principal energies to a file.

           Arguments:
            | ``temp`` -- The temperature to use for the temperature-dependent
                          quantities.
            | ``filename`` -- The name of the CSV file.
        """
        f = file(filename, "w")
        c = csv.writer(f)
        self.dump_table(temp, c)
        f.close()


    def dump_table(self, temp, c):
        """Write a CSV file with the principal energies to a stream .

           Arguments:
            | ``temp`` -- The temperature to use for the temperature-dependent
                          quantities.
            | ``c`` -- A csv.writer object from the built-in Python csv module.
        """
        c.writerow(["Temperature [K]", temp])
        c.writerow([])
        c.writerow(["Quantity"] +
                   [pf.title for pf, st in self._iter_pfs()] +
                   ["Linear combination (always in kJ/mol)"])
        c.writerow(["Signed stoichiometry"] +
                   [st for pf, st in self._iter_pfs()])
        c.writerow(["**Values in a.u.**"])
        c.writerow(["Electronic energy"] +
                   [pf.electronic.energy for pf, st in self._iter_pfs()] +
                   [self.energy_difference()/kjmol])
        c.writerow(["Zero-point energy"] +
                   [pf.zero_point_energy() for pf, st in self._iter_pfs()] +
                   [self.zero_point_energy_difference()/kjmol])
        c.writerow(["Internal heat (%.2fK)" % temp] +
                   [pf.internal_heat(temp) for pf, st in self._iter_pfs()] +
                   [self.internal_heat_difference(temp)/kjmol])
        c.writerow(["Chemical potential (%.2fK)" % temp] +
                   [pf.chemical_potential(temp) for pf, st in self._iter_pfs()] +
                   [self.free_energy_change(temp)/kjmol])
        c.writerow(["**Corrections in kJ/mol**"])
        c.writerow(["Zero-point energy"] +
                   [(pf.zero_point_energy() - pf.electronic.energy)/kjmol for pf, st in self._iter_pfs()] +
                   [(self.zero_point_energy_difference() - self.energy_difference())/kjmol])
        c.writerow(["Internal heat (%.2fK)" % temp] +
                   [(pf.internal_heat(temp) - pf.electronic.energy)/kjmol for pf, st in self._iter_pfs()] +
                   [(self.internal_heat_difference(temp) - self.energy_difference())/kjmol])
        c.writerow(["Chemical potential (%.2fK)" % temp] +
                   [(pf.chemical_potential(temp) - pf.electronic.energy)/kjmol for pf, st in self._iter_pfs()] +
                   [(self.free_energy_change(temp) - self.energy_difference())/kjmol])
        c.writerow([])
        c.writerow(["**Other quantities**", "Unit", "Value"])

    def write_to_file(self, filename):
        """Write the model to a text file.

           One argument:
            | ``filename`` -- The file to write the output.
        """
        f = file(filename, "w")
        self.dump(f)
        f.close()

    def dump(self, f):
        """Write all info about the model to a file."""
        print >> f, "The chemical balance:"
        print >> f, "  ", " + ".join("%s*(\"%s\")" % (-st, pf.title) for pf, st in self._iter_pfs() if st < 0),
        print >> f, " <--> ",
        print >> f, " + ".join("%s*(\"%s\")" % (st, pf.title) for pf, st in self._iter_pfs() if st > 0)
        print >> f
        for counter, (pf, st) in enumerate(self._iter_pfs()):
            print >> f, "Partition function %i" % counter
            print >> f, "Signed stoichiometry: %i" % st
            pf.dump(f)
            print >> f


class ThermodynamicModel(BaseModel):
    """A model for a thermodynamic equilibrium."""
    def __init__(self, pfs_react, pfs_prod):
        """
           Arguments:
            | ``pfs_react`` -- A list with reactant partition functions.
            | ``pfs_prod`` -- A list with product partition functions.

           Both arguments are lists whose items should be PartFun objects.
           One may also replace an item by a ``(pf, st)`` tuple, where ``pf`` is
           the partition function and ``st`` is the stoichiometry.

           Useful attributes:
            | ``unit_name`` -- A string with the SI unit of the equilibrium
                               constant
            | ``unit`` -- The conversion factor to transform the equilibrium
                          constant into SI units
                          (``equilibrium_const/self.unit``)
        """
        BaseModel.__init__(self)
        self._add_pfs(pfs_react, -1)
        self._add_pfs(pfs_prod, +1)
        self._set_unit()

    def dump_table(self, temp, c):
        """Write a CSV file with the principal energies to a stream .

           Arguments:
            | ``temp`` -- The temperature to use for the temperature-dependent
                          quantities.
            | ``c`` -- A csv.writer object from the built-in Python csv module.
        """
        BaseModel.dump_table(self, temp, c)
        c.writerow(["Equilibrium constant", self.unit_name, self.equilibrium_constant(temp)/self.unit])

    def dump(self, f):
        """Write all info about the thermodynamic model to a file."""
        delta_E = self.energy_difference()
        print >> f, "Electronic energy difference [kJ/mol] = %.1f" % (delta_E/kjmol)
        delta_ZPE = self.zero_point_energy_difference()
        print >> f, "Zero-point energy difference [kJ/mol] = %.1f" % (delta_ZPE/kjmol)
        print >> f
        BaseModel.dump(self, f)


class BaseKineticModel(BaseModel):
    """A generic model for the rate constant of a chemical reaction."""

    def rate_constant(self, temp, do_log=False):
        """Compute the rate constant of the reaction in this analysis

           Arguments:
            | ``temp`` -- The temperature.

           Optional argument:
            | ``do_log`` -- When True, the logarithm of the rate constant is
                            returned instead of just the rate constant itself.
                            [default=False]
        """
        raise NotImplementedError


class KineticModel(BaseKineticModel):
    """A model for the rate constant of a single-step chemical reaction."""
    def __init__(self, pfs_react, pf_trans, tunneling=None):
        """
           Arguments:
            | ``pfs_react`` -- A list of partition functions for the reactants.
            | ``pf_trans`` -- The partition function of the transition state.

           The first argument is a list whose items should be PartFun objects.
           One may also replace an item by a ``(pf, st)`` tuple, where ``pf`` is
           the partition function and ``st`` is the stoichiometry.

           Optional argument:
            | ``tunneling`` -- A tunneling correction object. If not given, no
                               tunneling correction is applied.


           Useful attributes:
            | ``unit_name`` -- A string containing the SI unit of the rate
                               constant.
            | ``unit`` -- The conversion factor to transform the rate constant
                          to SI units (``rate_const/self.unit``)
        """
        BaseKineticModel.__init__(self)
        if len(pfs_react) == 0:
            raise ValueError("At least one reactant must be given.")
        self._add_pfs(pfs_react, -1)
        self._add_pfs([pf_trans], +1)
        self.tunneling = tunneling
        self._set_unit(per_second=True)

    def rate_constant(self, temp, do_log=False):
        """See :meth:`BaseKineticModel.rate_constant`

           The implementation is based on transition state theory.
        """
        result = self.equilibrium_constant(temp, do_log)
        if do_log:
            result = np.log(boltzmann*temp/planck) + result
            if self.tunneling is not None:
                result += np.log(self.tunneling(temp))
        else:
            result = boltzmann*temp/planck*result
            if self.tunneling is not None:
                result *= self.tunneling(temp)
        return result

    def dump_table(self, temp, c):
        """Write a CSV file with the principal energies to a stream .

           Arguments:
            | ``temp`` -- The temperature to use for the temperature-dependent
                          quantities.
            | ``c`` -- A csv.writer object from the built-in Python csv module.
        """
        BaseKineticModel.dump_table(self, temp, c)
        c.writerow(["Rate constant", self.unit_name, self.rate_constant(temp)/self.unit])

    def dump(self, f):
        """Write all info about the kinetic model to a file."""
        delta_E = self.energy_difference()
        print >> f, "Electronic energy barrier [kJ/mol] = %.1f" % (delta_E/kjmol)
        delta_ZPE = self.zero_point_energy_difference()
        print >> f, "Zero-point energy barrier [kJ/mol] = %.1f" % (delta_ZPE/kjmol)
        print >> f
        BaseKineticModel.dump(self, f)


class ActivationKineticModel(BaseKineticModel):
    """A model for the rate constant of a single-step chemical reaction with a pre-reactive complex."""
    def __init__(self, tm, km):
        """
           Arguments:
            | ``tm`` -- The thermodynamic model for the pre-reactive complex.
            | ``km`` -- The kinetic model for the single-step reaction.
        """
        BaseKineticModel.__init__(self)
        self.tm = tm
        self.km = km
        self._add_pfs(tm.pfs_all.iteritems())
        self._add_pfs(km.pfs_all.iteritems())
        self._set_unit(per_second=True)

    def rate_constant(self, temp, do_log=False):
        """See :meth:`BaseKineticModel.rate_constant`"""
        if do_log:
            return self.tm.equilibrium_constant(temp, True) + \
                   self.km.rate_constant(temp, True)
        else:
            return self.tm.equilibrium_constant(temp, False)* \
                   self.km.rate_constant(temp, False)

    def dump(self, f):
        """Write all info about the kinetic model to a file."""
        delta_E = self.energy_difference()
        print >> f, "Electronic energy barrier [kJ/mol] = %.1f" % (delta_E/kjmol)
        delta_ZPE = self.zero_point_energy_difference()
        print >> f, "Zero-point energy barrier [kJ/mol] = %.1f" % (delta_ZPE/kjmol)
        print >> f
        print >> f, "Thermodynamic submodel"
        self.tm.dump(f)
        print >> f
        print >> f, "Kinetic submodel"
        self.km.dump(f)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# TAMkin is a post-processing toolkit for normal mode analysis, thermochemistry
# and reaction kinetics.
# Copyright (C) 2008-2012 Toon Verstraelen <Toon.Verstraelen@UGent.be>, An Ghysels
# <An.Ghysels@UGent.be> and Matthias Vandichel <Matthias.Vandichel@UGent.be>
# Center for Molecular Modeling (CMM), Ghent University, Ghent, Belgium; all
# rights reserved unless otherwise stated.
#
# This file is part of TAMkin.
#
# TAMkin is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# In addition to the regulations of the GNU General Public License,
# publications and communications based in parts on this program or on
# parts of this program are required to cite the following article:
#
# "TAMkin: A Versatile Package for Vibrational Analysis and Chemical Kinetics",
# An Ghysels, Toon Verstraelen, Karen Hemelsoet, Michel Waroquier and Veronique
# Van Speybroeck, Journal of Chemical Information and Modeling, 2010, 50,
# 1736-1750W
# http://dx.doi.org/10.1021/ci100099g
#
# TAMkin is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
#--


import glob, os
from optparse import OptionParser

from states import *

usage = """
%prog create lot_label basis_label
%prog clone geo_lot_label geo_basis_label sp_lot_label sp_basis_label
"""

def main():
    parser = OptionParser(usage)
    parser.add_option(
        "-r", "--random", default=False, action='store_true',
        help="Randomly distort the initial geometries with uniform "
             "displacements of [-0.1*angstrom,0.1*angstrom] in x, y, and z "
             "directions."
    )
    parser.add_option(
        "-s", "--suffix", default="",
        help="Append __SUFFIX to the directory name"
    )
    options, args = parser.parse_args()

    if len(args) == 0:
        parser.error("Expecting at least one argument")

    if args[0] == "create":
        if len(args) != 3:
            parser.error("create needs two extra arguments")
        lot_label, basis_label = args[1:]
        root = get_root(lot_label, basis_label, options.suffix)
        for state in states:
            for job in state.jobs:
                if job.name == "sp":
                    continue
                job.write_input(state, root, lot_label, basis_label, options.suffix, options.random)
    elif args[0] == "clone":
        if len(args) != 5:
            parser.error("create needs four extra arguments")
        lot0_label, basis0_label, lot1_label, basis1_label = args[1:]
        root0 = get_root(lot0_label, basis0_label, options.suffix)
        root1 = "GEO__%s__ENERGY__%s" % (
            get_root(lot0_label, basis0_label, ""),
            get_root(lot1_label, basis1_label, options.suffix)
        )
        if not os.path.isdir(root1):
            os.mkdir(root1)
        for dirname in glob.glob("%s/*" % root0):
            if os.path.isdir(dirname) and not (dirname.endswith("sp") or dirname.endswith("bsse")):
                jobdir = os.path.basename(dirname)
                source = "../%s/%s" % (root0, jobdir)
                destination = "%s/%s" % (root1, jobdir)
                if not os.path.exists(destination):
                    os.symlink(source, destination)
        for state in states:
            for job in state.jobs:
                if job.name == "bsse" or job.name == "sp" or job.name.startswith("cps"):
                    job.write_input(state, root1, lot1_label, basis1_label, options.suffix, options.random)
    else:
        parser.error("Unknown command: %s" % args[0])



if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# TAMkin is a post-processing toolkit for normal mode analysis, thermochemistry
# and reaction kinetics.
# Copyright (C) 2008-2012 Toon Verstraelen <Toon.Verstraelen@UGent.be>, An Ghysels
# <An.Ghysels@UGent.be> and Matthias Vandichel <Matthias.Vandichel@UGent.be>
# Center for Molecular Modeling (CMM), Ghent University, Ghent, Belgium; all
# rights reserved unless otherwise stated.
#
# This file is part of TAMkin.
#
# TAMkin is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# In addition to the regulations of the GNU General Public License,
# publications and communications based in parts on this program or on
# parts of this program are required to cite the following article:
#
# "TAMkin: A Versatile Package for Vibrational Analysis and Chemical Kinetics",
# An Ghysels, Toon Verstraelen, Karen Hemelsoet, Michel Waroquier and Veronique
# Van Speybroeck, Journal of Chemical Information and Modeling, 2010, 50,
# 1736-1750W
# http://dx.doi.org/10.1021/ci100099g
#
# TAMkin is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
#--
#!/usr/bin/env python

# This script takes a list of Gaussian fchk files as arguments and will strip
# all sections that are irrelevant for TAMkin. Also the trailing white space on
# each line is stripped. This is used to reduce the size of the example fchk
# files in TAMkin.


keep_fields = set([
    "Multiplicity", "Total Energy", "Atomic numbers",
    "Current cartesian coordinates", "Real atomic weights",
    "Cartesian Gradient", "Cartesian Force Constants",
])

def strip(fn):
    # a list with all lines that we'd like to keep
    lines = []
    # load the good lines
    f = file(fn, "r")
    busy = False
    keep = False
    for line in f:
        line = line.rstrip()
        if len(lines) < 2:
            # keep the first two lines
            lines.append(line)
        else:
            if busy:
                if not line.startswith(" "):
                    keep = False
                    busy = False
                if keep:
                    lines.append(line)
            if not busy:
                busy = line[47:49] == "N="
                field = line[:43].strip()
                if field in keep_fields:
                    lines.append(line)
                    keep = True
                else:
                    keep = False
    f.close()
    # print stuff back into the same file
    f = file(fn, "w")
    for line in lines:
        print >> f, line
    f.close()


if __name__ == "__main__":
    import sys
    fns_fchk = sys.argv[1:]
    for fn in fns_fchk:
        if fn.endswith(".fchk"):
            print "Stripping", fn
            strip(fn)
        else:
            print "Skipping", fn, "(wrong extension)"



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# TAMkin is a post-processing toolkit for normal mode analysis, thermochemistry
# and reaction kinetics.
# Copyright (C) 2008-2012 Toon Verstraelen <Toon.Verstraelen@UGent.be>, An Ghysels
# <An.Ghysels@UGent.be> and Matthias Vandichel <Matthias.Vandichel@UGent.be>
# Center for Molecular Modeling (CMM), Ghent University, Ghent, Belgium; all
# rights reserved unless otherwise stated.
#
# This file is part of TAMkin.
#
# TAMkin is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# In addition to the regulations of the GNU General Public License,
# publications and communications based in parts on this program or on
# parts of this program are required to cite the following article:
#
# "TAMkin: A Versatile Package for Vibrational Analysis and Chemical Kinetics",
# An Ghysels, Toon Verstraelen, Karen Hemelsoet, Michel Waroquier and Veronique
# Van Speybroeck, Journal of Chemical Information and Modeling, 2010, 50,
# 1736-1750W
# http://dx.doi.org/10.1021/ci100099g
#
# TAMkin is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --


from __future__ import print_function

import os
import subprocess
import sys

from setuptools import setup


pypkg="tamkin"


# Try to get the version from git describe
__version__ = None
try:
    print('Trying to get the version from git describe')
    git_describe = subprocess.check_output(["git", "describe", "--tags"])
    version_words = git_describe.decode('utf-8').strip().split('-')
    __version__ = version_words[0]
    if len(version_words) > 1:
        __version__ += '.post' + version_words[1]
    print('Version from git describe: {}'.format(__version__))
except (subprocess.CalledProcessError, OSError):
    pass

# Interact with version.py
fn_version = os.path.join(os.path.dirname(__file__), pypkg, 'version.py')
version_template = """\
\"""Do not edit this file, versioning is governed by ``git describe --tags`` and ``setup.py``.\"""
__version__ = '{}'
"""
if __version__ is None:
    print('Trying to get the version from {}',format(fn_version))
    # Try to load the git version tag from version.py
    try:
        with open(fn_version, 'r') as fh:
            __version__ = fh.read().split('=')[-1].replace('\'', '').strip()
    except IOError:
        print('Could not determine version. Giving up.')
        sys.exit(1)
    print('Version according to {}: {}'.format(fn_version, __version__))
else:
    # Store the git version tag in version.py
    print('Writing version to {}'.format(fn_version))
    with open(fn_version, 'w') as fh:
        fh.write(version_template.format(__version__))


setup(
    name=pypkg,
    version=__version__,
    description='TAMkin is a post-processing toolkit for thermochemistry and kinetics analysis.',
    author='Toon Verstraelen, Matthias Vandichel, An Ghysels',
    author_email='Toon.Verstraelen@UGent.be, Matthias.Vandichel@UGent.be, An.Ghysels@UGent.be',
    url='http://molmod.ugent.be/code/',
    package_dir = {'tamkin': 'tamkin'},
    packages = ['tamkin', 'tamkin.io', 'tamkin.test'],
    scripts=["scripts/tamkin-driver"],
    include_package_data=True,
    install_requires=['numpy>=1.0', 'nose>=0.11', 'matplotlib>1.1',
                      'molmod>=1.3.2', 'scipy>=0.17.1'],
    classifiers=[
        'Environment :: Console',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 2.7',
        'Topic :: Scientific/Engineering :: Physics',
        'Topic :: Scientific/Engineering :: Chemistry',
        'Intended Audience :: Science/Research',
    ])



# -*- coding: utf-8 -*-
# TAMkin is a post-processing toolkit for normal mode analysis, thermochemistry
# and reaction kinetics.
# Copyright (C) 2008-2012 Toon Verstraelen <Toon.Verstraelen@UGent.be>, An Ghysels
# <An.Ghysels@UGent.be> and Matthias Vandichel <Matthias.Vandichel@UGent.be>
# Center for Molecular Modeling (CMM), Ghent University, Ghent, Belgium; all
# rights reserved unless otherwise stated.
#
# This file is part of TAMkin.
#
# TAMkin is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# In addition to the regulations of the GNU General Public License,
# publications and communications based in parts on this program or on
# parts of this program are required to cite the following article:
#
# "TAMkin: A Versatile Package for Vibrational Analysis and Chemical Kinetics",
# An Ghysels, Toon Verstraelen, Karen Hemelsoet, Michel Waroquier and Veronique
# Van Speybroeck, Journal of Chemical Information and Modeling, 2010, 50,
# 1736-1750W
# http://dx.doi.org/10.1021/ci100099g
#
# TAMkin is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
"""Tunneling models to approximate QM behavior in reaction coordinate.

   Instances of the classes :class:`Eckart` or :class:`Wigner` act as functions
   that take one argument, the temperature, and return a correction factor
   for the rate coefficient. Such an object can be given as an optional
   argument to the constructor of a :class:`tamkin.chemmod.KineticModel`
   object to take tunneling corrections into account for the computation of
   rate constants.
"""


from molmod import boltzmann, planck, kjmol, lightspeed, centimeter

import numpy as np


__all__ = ["TunnelingCorrection", "Eckart", "Wigner", "Miller"]


class TunnelingCorrection(object):
    """Abstract base class for the implementation of a Tunneling correction

       This class merely defines the interface and holds some docstrings.
    """
    def __call__(self, temps):
        """Compute a the tunneling correction as function of the temperature

           Argument:
            | ``temps`` -- a numpy array of temperatures or a single temperature

           Derived classes must override this method with a function that
           computes the correction factors for the rate constant at the given
           temperatures.
        """
        raise NotImplementedError

    def dump(self, f):
        '''Write descriptive info about the tunneling correction to a text file'''
        raise NotImplementedError


class Eckart(TunnelingCorrection):
    """Implements the Eckart tunneling correction factor

       This correction is proposed in C. Eckart, Phys. Rev. 35, 1303 (1930),
       http://link.aps.org/doi/10.1103/PhysRev.35.1303.
    """

    def __init__(self, pfs_react, pf_trans, pfs_prod):
        """
           Arguments:
            | ``pfs_react`` -- a list with partition functions of the reactants
            | ``pf_trans`` -- the partition function of the transition state
            | ``pfs_prod`` -- a list with partition functions of the products

           Attributes derived from these arguments:
            | ``self.Ef`` -- forward energy barrier
            | ``self.Er`` -- reverse energy barrier
            | ``self.nu`` -- the imaginary frequency (as a real number)

           Note that this correction is only defined for transition states
           with only one imaginary frequency.
        """
        if len(pf_trans.vibrational.negative_freqs) != 1:
            raise ValueError("The partition function of the transition state must have exactly one negative frequency, found %i" % len(pf_trans.negative_freqs))
        if len(pfs_react) == 0:
            raise ValueError("At least one reactant is required.")
        if len(pfs_react) == 0:
            raise ValueError("At least one product is required.")
        self.Ef = pf_trans.electronic.energy - sum(pf.electronic.energy for pf in pfs_react)
        if self.Ef < 0:
            raise ValueError("The forward barrier is negative. Can not apply Eckart tunneling.")
        self.Er = pf_trans.electronic.energy - sum(pf.electronic.energy for pf in pfs_prod)
        if self.Er < 0:
            raise ValueError("The reverse barrier is negative. Can not apply Eckart tunneling.")
        self.nu = pf_trans.vibrational.negative_freqs[0]

    @classmethod
    def _from_parameters(cls, Ef, Er, nu):
        """An alternative constructor used for testing purposes.

           Arguments:
            | ``Ef`` -- The forward barrier
            | ``Er`` -- The reverse barrier
            | ``nu`` -- The imaginary frequency (as a real number)

           The constructor should not be used in normal situations.
        """
        if Ef < 0:
            raise ValueError("The forward barrier is negative. Can not apply Eckart tunneling.")
        if Er < 0:
            raise ValueError("The reverse barrier is negative. Can not apply Eckart tunneling.")
        result = cls.__new__(cls)
        result.Ef = Ef
        result.Er = Er
        result.nu = nu
        return result

    def dump(self, f):
        '''Write descriptive info about the tunneling correction to a text file'''
        print >> f, 'Eckart Tunneling Correction'
        print >> f, 'Forward Barrier [kJ/mol] = %.2f' % (self.Ef/kjmol)
        print >> f, 'Reverse Barrier [kJ/mol] = %.2f' % (self.Er/kjmol)
        print >> f, 'Imaginary frequeny [cm^-1] = %.1f' % (self.nu/(lightspeed/centimeter))
        print >> f

    def _compute_one_temp(self, temp):
        """Computes the correction for one temperature

           Arguments:
            | ``temp`` -- the temperature (scalar)
        """
        from scipy.integrate import quad

        h = 2*np.pi # the Planck constant in atomic units
        l = (self.Ef**(-0.5) + self.Er**(-0.5))**(-1)*np.sqrt(2) / self.nu

        def alpha(E):
            return np.sqrt(2*l**2*E/h**2)

        def beta(E):
            return np.sqrt(2*l**2*(E -(self.Ef-self.Er))/h**2)

        def delta(E):
            return np.sqrt(4*self.Ef*self.Er/(h*self.nu)**2-0.25)

        def P(E):
            return (
                np.cosh(2*np.pi*(alpha(E) + beta(E))) -
                np.cosh(2*np.pi*(alpha(E) - beta(E)))
            ) / (
                np.cosh(2*np.pi*(alpha(E) + beta(E))) +
                np.cosh(2*np.pi*(delta(E)))
            )

        def integrandum(E):
            return P(E)*np.exp(-(E-self.Ef)/(boltzmann*temp))

        # integration interval
        emin = max([0, self.Ef-self.Er])
        emax = 500*kjmol # The maximum reasonable barrier height.

        # this is just a sanity check to see whether the integrandum is
        # negligible at the borders of the integration interval
        energies = np.arange(emin, emax, 1*kjmol)
        integranda = np.array([integrandum(energy) for energy in energies])
        if max(integranda) * 1e-5 < max([integranda[0], integranda[-1]]):
            print "Integrandum is not negligible at borders.", integranda[0] / max(integranda), integranda[-1] / max(integranda)

        # actual integration
        integral, error = quad(integrandum, emin, emax)
        factor = 1.0/(boltzmann*temp)
        return integral*factor

    def __call__(self, temps):
        """See :meth:`TunnelingCorrection.__call__`."""
        if hasattr(temps, "__len__"):
            result = np.zeros(len(temps))
            for i, temp in enumerate(temps):
                result[i] = self._compute_one_temp(temp)
            return result
        else:
            return self._compute_one_temp(temps)


class Wigner(TunnelingCorrection):
    """Implements the Wigner tunneling correction factor

       This correction is proposed in E. Wigner, Z. Physik. Chern. B 19, 203
       (1932).
    """
    def __init__(self, pf_trans):
        """
           Arguments:
            | ``pf_trans`` -- the partition function of the transition state

           Attribute derived from these argument:
            | ``self.nu`` -- the imaginary frequency (as a real number)

           Note that this correction is only defined for transition states
           with only one imaginary frequency.
        """
        if len(pf_trans.vibrational.negative_freqs) != 1:
            raise ValueError("The partition function of the transition state must have exactly one negative frequency, found %i" % len(pf_trans.negative_freqs))
        self.nu = pf_trans.vibrational.negative_freqs[0]

    @classmethod
    def _from_parameters(cls, nu):
        """An alternative constructor used for testing purposes.

           Arguments:
            | ``nu`` -- The imaginary frequency (as a real number)

           This method should not be used in normal situations.
        """
        result = cls.__new__(cls)
        result.nu = nu
        return result

    def dump(self, f):
        '''Write descriptive info about the tunneling correction to a text file'''
        print >> f, 'Wigner Tunneling Correction'
        print >> f, 'Imaginary frequeny [cm^-1] = %.1f' % (self.nu/(lightspeed/centimeter))
        print >> f

    def __call__(self, temps):
        """See :meth:`TunnelingCorrection.__call__`."""
        return 1+(planck*self.nu/(boltzmann*temps))**2/24


class Miller(TunnelingCorrection):
    """Implements the Miller tunneling correction factor

       This correction is proposed in Miller, W. H. J. Chem. Phys. 1973, 61,
       1823, http://dx.doi.org/10.1063/1.1682181.
    """
    def __init__(self, pf_trans):
        """
           Arguments:
            | ``pf_trans`` -- the partition function of the transition state

           Attribute derived from these argument:
            | ``self.nu`` -- the imaginary frequency (as a real number)

           Note that this correction is only defined for transition states
           with only one imaginary frequency.
        """
        if len(pf_trans.vibrational.negative_freqs) != 1:
            raise ValueError("The partition function of the transition state must have exactly one negative frequency, found %i" % len(pf_trans.negative_freqs))
        self.nu = pf_trans.vibrational.negative_freqs[0]

    @classmethod
    def _from_parameters(cls, nu):
        """An alternative constructor used for testing purposes.

           Arguments:
            | ``nu`` -- The imaginary frequency (as a real number)

           This method should not be used in normal situations.
        """
        result = cls.__new__(cls)
        result.nu = nu
        return result

    def dump(self, f):
        '''Write descriptive info about the tunneling correction to a text file'''
        print >> f, 'Miller Tunneling Correction'
        print >> f, 'Imaginary frequeny [cm^-1] = %.1f' % (self.nu/(lightspeed/centimeter))
        print >> f

    def __call__(self, temps):
        """See :meth:`TunnelingCorrection.__call__`."""
        x = 0.5*planck*self.nu/(boltzmann*temps)
        return x/np.sin(x)



# -*- coding: utf-8 -*-
# TAMkin is a post-processing toolkit for normal mode analysis, thermochemistry
# and reaction kinetics.
# Copyright (C) 2008-2012 Toon Verstraelen <Toon.Verstraelen@UGent.be>, An Ghysels
# <An.Ghysels@UGent.be> and Matthias Vandichel <Matthias.Vandichel@UGent.be>
# Center for Molecular Modeling (CMM), Ghent University, Ghent, Belgium; all
# rights reserved unless otherwise stated.
#
# This file is part of TAMkin.
#
# TAMkin is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# In addition to the regulations of the GNU General Public License,
# publications and communications based in parts on this program or on
# parts of this program are required to cite the following article:
#
# "TAMkin: A Versatile Package for Vibrational Analysis and Chemical Kinetics",
# An Ghysels, Toon Verstraelen, Karen Hemelsoet, Michel Waroquier and Veronique
# Van Speybroeck, Journal of Chemical Information and Modeling, 2010, 50,
# 1736-1750W
# http://dx.doi.org/10.1021/ci100099g
#
# TAMkin is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --


import os
import numpy as np
import pkg_resources
import unittest

from molmod import lightspeed, centimeter
from molmod.test.common import tmpdir

from tamkin import *


__all__ = ["NMAToolsTestCase"]


class NMAToolsTestCase(unittest.TestCase):

    def test_load_coordinates_charmm(self):
        molecule = load_molecule_charmm(
            pkg_resources.resource_filename(__name__, "../data/test/an/ethanol.cor"),
            pkg_resources.resource_filename(__name__, "../data/test/an/ethanol.hess.full"))
        coordinates, masses, symbols = load_coordinates_charmm(
            pkg_resources.resource_filename(__name__, "../data/test/an/ethanol.cor"))
        for at in range(9):
            self.assertAlmostEqual(molecule.masses[at], masses[at], 3)
            for j in range(3):
                self.assertAlmostEqual(molecule.coordinates[at,j], coordinates[at,j], 3)

    def test_load_modes_charmm(self):
        molecule = load_molecule_charmm(
            pkg_resources.resource_filename(__name__, "../data/test/an/ethanol.cor"),
            pkg_resources.resource_filename(__name__, "../data/test/an/ethanol.hess.full"))

        # full Hessian
        nma = NMA(molecule)
        modes2, freqs2, masses2 = load_modes_charmm(
            pkg_resources.resource_filename(__name__, "../data/test/an/ethanol.modes.full"))
        for index in range(6,27):
            for j in range(27):
                self.assertAlmostEqual(abs(nma.modes[j,index]), abs(modes2[j,index]), 7)
        for at in range(27):
            self.assertAlmostEqual(nma.freqs[at], freqs2[at], 7)

        # MBH
        nma = NMA(molecule, MBH([[5,6,7,8]]))
        modes2, freqs2, masses2 = load_modes_charmm(
            pkg_resources.resource_filename(__name__, "../data/test/an/ethanol.modes.mbh"))
        for index in range(6,21):
            for j in range(27):
                self.assertAlmostEqual(abs(nma.modes[j,index]), abs(modes2[j,index]), 7)
        for at in range(21):
            self.assertAlmostEqual(nma.freqs[at], freqs2[at], 7)

    def test_overlap(self):
        molecule = load_molecule_charmm(
            pkg_resources.resource_filename(__name__, "../data/test/an/ethanol.cor"),
            pkg_resources.resource_filename(__name__, "../data/test/an/ethanol.hess.full"))
        nma1 = NMA(molecule)
        fixed = load_indices(
            pkg_resources.resource_filename(__name__, "../data/test/an/fixed.06.txt"))
        nma2 = NMA(molecule, PHVA(fixed))
        overlap = compute_overlap(nma1, nma2)
        overlap = compute_overlap((nma1.modes, nma1.freqs), (nma2.modes, nma2.freqs))
        overlap = compute_overlap(nma1.modes, nma2.modes)
        overlap = compute_overlap(nma1.modes[:,0], nma2.modes[:,0])
        # TODO
        #self.assertAlmostEqual()

    def test_delta_vector(self):
        # from charmmcor
        coor1,masses1,symb1 = load_coordinates_charmm(
            pkg_resources.resource_filename(__name__, "../data/test/an/ethanol.cor"))
        coor2,masses2,symb2 = load_coordinates_charmm(
            pkg_resources.resource_filename(__name__, "../data/test/an/ethanol.2.cor"))
        delta = compute_delta(coor1, coor2)
        # TODO
        #self.assertAlmostEqual()
        delta = compute_delta(coor1, coor2, masses=masses1, normalize=True)
        #self.assertAlmostEqual()

    def test_eigenvalue_sensitivity(self):
        molecule = load_molecule_charmm(
            pkg_resources.resource_filename(__name__, "../data/test/an/ethanol.cor"),
            pkg_resources.resource_filename(__name__, "../data/test/an/ethanol.hess.full"))
        nma = NMA(molecule)
        for i in range(7,27):
            sensit = compute_sensitivity_freq(nma, i)
            self.assertAlmostEqual(np.sum((np.dot(sensit,nma.modes)-nma.modes)**2,0)[i],0.0,9)

    def test_create_blocks_peptide_charmm(self):
        blocks1 = create_blocks_peptide_charmm(
            pkg_resources.resource_filename(__name__, "../data/test/charmm/crambin.crd"),
            "RTB", blocksize=1)
        blocks2 = create_blocks_peptide_charmm(
            pkg_resources.resource_filename(__name__, "../data/test/charmm/crambin.crd"),
            "RTB", blocksize=2)
        self.assertEqual(len(blocks1)/2+1, len(blocks2))
        subs1 = create_subs_peptide_charmm(
            pkg_resources.resource_filename(__name__, "../data/test/charmm/crambin.crd"),
            frequency=1)
        subs2 = create_subs_peptide_charmm(
            pkg_resources.resource_filename(__name__, "../data/test/charmm/crambin.crd"),
            frequency=2)
        self.assertEqual(len(subs1)/2, len(subs2))

        blocks = create_blocks_peptide_charmm(
            pkg_resources.resource_filename(__name__, "../data/test/charmm/crambin.crd"))
        self.assertEqual(len(blocks), 91)
        blocks = create_blocks_peptide_charmm(
            pkg_resources.resource_filename(__name__, "../data/test/charmm/crambin.crd"),
            "dihedral")
        self.assertEqual(len(blocks), 91)
        blocks = create_blocks_peptide_charmm(
            pkg_resources.resource_filename(__name__, "../data/test/charmm/crambin.crd"),
            "RHbending")
        self.assertEqual(len(blocks),136)
        blocks = create_blocks_peptide_charmm(
            pkg_resources.resource_filename(__name__, "../data/test/charmm/crambin.crd"),
            "normal")
        self.assertEqual(len(blocks), 91)

    def test_plot_spectrum(self):
        molecule = load_molecule_charmm(
            pkg_resources.resource_filename(__name__, "../data/test/an/ethanol.cor"),
            pkg_resources.resource_filename(__name__, "../data/test/an/ethanol.hess.full"))
        nma = NMA(molecule)
        invcm = lightspeed/centimeter
        with tmpdir(__name__, 'test_plot_spectrum') as dn:
            plot_spectrum_lines(
                os.path.join(dn, "spectrum-lines.1.png"),
                [nma.freqs, nma.freqs], title="standard settings")
            plot_spectrum_lines(
                os.path.join(dn, "spectrum-lines.2.png"),
                [nma.freqs, nma.freqs], low=-10.0*invcm, high=500.0*invcm, title="zoom")

            plot_spectrum_dos(
                os.path.join(dn, "spectrum-dos.1.png"),
                [nma.freqs], title="standard settings")
            plot_spectrum_dos(
                os.path.join(dn, "spectrum-dos.2.png"),
                [nma.freqs], low=-10.0*invcm, high=1500.0*invcm, title="zoom")
            plot_spectrum_dos(
                os.path.join(dn, "spectrum-dos.3.png"),
                [nma.freqs], low=-10.0*invcm, high=1500.0*invcm, width=50.0*invcm, title="width")
            plot_spectrum_dos(
                os.path.join(dn, "spectrum-dos.4.png"),
                [nma.freqs], low=-10.0*invcm, high=1500.0*invcm, width=50.0*invcm, step=20.0*invcm, title="step size")
            plot_spectrum_dos(
                os.path.join(dn, "spectrum-dos.5.png"),
                [nma.freqs, nma.freqs*1.1], title="two spectra")
            plot_spectrum_dos(
                os.path.join(dn, "spectrum-dos.6.png"),
                [nma.freqs, nma.freqs*1.1], all_amps=[1.0,2.0], title="different amplitude")

    def test_create_enm_molecule(self):
        molecule = load_molecule_charmm(
            pkg_resources.resource_filename(__name__, "../data/test/an/ethanol.cor"),
            pkg_resources.resource_filename(__name__, "../data/test/an/ethanol.hess.full"))
        selected = range(5)
        mol = create_enm_molecule(molecule, selected)
        nma = NMA(mol)

        mol = create_enm_molecule(molecule.coordinates, rcut=5)
        nma = NMA(mol)

        mol = create_enm_molecule(molecule.coordinates, selected, masses=np.ones(molecule.size)*2.0, rcut=5)
        nma = NMA(mol)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# TAMkin is a post-processing toolkit for normal mode analysis, thermochemistry
# and reaction kinetics.
# Copyright (C) 2008-2012 Toon Verstraelen <Toon.Verstraelen@UGent.be>, An Ghysels
# <An.Ghysels@UGent.be> and Matthias Vandichel <Matthias.Vandichel@UGent.be>
# Center for Molecular Modeling (CMM), Ghent University, Ghent, Belgium; all
# rights reserved unless otherwise stated.
#
# This file is part of TAMkin.
#
# TAMkin is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# In addition to the regulations of the GNU General Public License,
# publications and communications based in parts on this program or on
# parts of this program are required to cite the following article:
#
# "TAMkin: A Versatile Package for Vibrational Analysis and Chemical Kinetics",
# An Ghysels, Toon Verstraelen, Karen Hemelsoet, Michel Waroquier and Veronique
# Van Speybroeck, Journal of Chemical Information and Modeling, 2010, 50,
# 1736-1750W
# http://dx.doi.org/10.1021/ci100099g
#
# TAMkin is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
#--


from lot_basis import lots_list
import html

def load_summary(fn):
    f = file(fn)
    result = tuple(float(word) for word in f.next().split())
    f.close()
    return result[:-4], result[-4:-2], result[-2:]


def ts_bsse(fn_template, title, rows):
    if len(rows) == 0:
        rows.append([u"<th>Approach→<br />Functional↓</th>"])
        rows.append(["<th></th>"])
    rows[0].append("<th colspan=3>%s</th>" % title.replace(", ", "<br />"))
    rows[1].append(u"<td>ΔE</td><td>ΔE'</td><td>ΔE'-ΔE</td>")
    counter = 2
    for lot in lots_list:
        if lot.spin == "ROS":
            lot_label = "ro" + lot.label
        else:
            lot_label = lot.label
        if len(rows) <= counter:
            rows.append(["<th>%s</th>" % lot_label])
        try:
            (A, Ea), (dE0, dE) = load_summary(fn_template % (lot_label, "bss"))[1:]
            (Ap, Eap), (dE0p, dEp) = load_summary(fn_template % (lot_label, "cps"))[1:]
            rows[counter].append("<td>%.0f</td>" % dE)
            rows[counter].append("<td>%.0f</td>" % dEp)
            rows[counter].append("<td style='background-color:#DDD;'>%.0f</td>" % (dEp-dE))
            #rows[counter].append("<td>%.0f</td>" % Ea)
            #rows[counter].append("<td>%.0f</td>" % Eap)
            #rows[counter].append("<td style='background-color:#DDD;'>%.0f</td>" % (Eap-Ea))
        except IOError, StopIteration:
            #rows[counter].append("<td>&nbsp</td><td>&nbsp</td><td>&nbsp</td>")
            rows[counter].append("<td>&nbsp</td><td>&nbsp</td><td>&nbsp</td>")
        counter += 1

f = file("bssetab.html", "w")
print >> f, html.header % "BSSE Overview"

rows = []
ts_bsse("%s__6-31gd/ho_%s_summary_gauche.txt", "Gauche, HO, Consistent, 6-31G(d)", rows)
ts_bsse("%s__6-31gd/ho_%s_summary_trans.txt", "Trans, HO, Consistent, 6-31G(d)", rows)
ts_bsse("%s__6-311+g3df2p/ho_%s_summary_gauche.txt", "Gauche, HO, Consistent, 6-311+G(3df,2p)", rows)
ts_bsse("%s__6-311+g3df2p/ho_%s_summary_trans.txt", "Trans, HO, Consistent, 6-311+G(3df,2p)", rows)
ts_bsse("GEO__b3lyp__6-31gd__ENERGY__%s__6-311+g3df2p/ho_%s_summary_gauche.txt", "Gauche, HO, GEO=B3LYP/6-31G(d), 6-311+G(3df,2p)", rows)
ts_bsse("GEO__b3lyp__6-31gd__ENERGY__%s__6-311+g3df2p/ho_%s_summary_trans.txt", "Trans, HO, GEO=B3LYP/6-31G(d), 6-311+G(3df,2p)", rows)

print >> f, "<p>BSSE corrections on the transition state.</p>"
html.print_table(f, rows)

print >> f, html.footer
